# Copyright (c) 2025 NetApp, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock
from unittest.mock import patch
import uuid

import ddt

from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap.client import (
    fakes as fake_client)
from cinder.tests.unit.volume.drivers.netapp.dataontap import fakes as fake
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp.dataontap.client import client_cmode
from cinder.volume.drivers.netapp.dataontap.client import client_cmode_rest
from cinder.volume.drivers.netapp.dataontap.client.client_cmode_rest_asar2\
    import RestClientASAr2
from cinder.volume.drivers.netapp import utils as netapp_utils


CONNECTION_INFO = {'hostname': 'hostname',
                   'transport_type': 'https',
                   'port': 443,
                   'username': 'admin',
                   'password': 'passw0rd',
                   'vserver': 'fake_vserver',
                   'ssl_cert_path': 'fake_ca',
                   'api_trace_pattern': 'fake_regex',
                   'private_key_file': 'fake_private_key.pem',
                   'certificate_file': 'fake_cert.pem',
                   'ca_certificate_file': 'fake_ca_cert.crt',
                   'certificate_host_validation': 'False',
                   'is_disaggregated': 'True',  # ASA r2 is disaggregated
                   }


@ddt.ddt
class NetAppRestCmodeASAr2ClientTestCase(test.TestCase):

    def setUp(self):
        super(NetAppRestCmodeASAr2ClientTestCase, self).setUp()

        # Setup Client mocks
        self.mock_object(client_cmode.Client, '_init_ssh_client')
        # store the original reference so we can call it later in
        # test__get_cluster_nodes_info
        self.original_get_cluster_nodes_info = (
            client_cmode.Client._get_cluster_nodes_info)
        self.mock_object(client_cmode.Client, '_get_cluster_nodes_info',
                         return_value=fake.HYBRID_SYSTEM_NODES_INFO)
        self.mock_object(client_cmode.Client, 'get_ontap_version',
                         return_value=(9, 16, 1))
        self.mock_object(client_cmode.Client,
                         'get_ontapi_version',
                         return_value=(0, 0))

        # Setup RestClient mocks
        self.mock_object(client_cmode_rest.RestClient, '_init_ssh_client')

        self.original_get_cluster_nodes_info = (
            client_cmode_rest.RestClient._get_cluster_nodes_info)

        if not hasattr(client_cmode_rest.RestClient,
                       '_get_cluster_nodes_info'):
            setattr(client_cmode_rest.RestClient,
                    '_get_cluster_nodes_info',
                    None)
        self.original_get_cluster_nodes_info = (
            client_cmode_rest.RestClient._get_cluster_nodes_info)

        self.mock_object(client_cmode_rest.RestClient,
                         '_get_cluster_nodes_info',
                         return_value=fake.HYBRID_SYSTEM_NODES_INFO)
        self.mock_object(client_cmode_rest.RestClient, 'get_ontap_version',
                         return_value=(9, 16, 1))

        # Setup ASA r2 specific mocks
        self.mock_object(RestClientASAr2, '_init_ssh_client')
        self.mock_object(RestClientASAr2, '_get_cluster_nodes_info',
                         return_value=fake.HYBRID_SYSTEM_NODES_INFO)
        self.mock_object(RestClientASAr2, 'get_ontap_version',
                         return_value=(9, 16, 1))

        with mock.patch.object(RestClientASAr2,
                               'get_ontap_version',
                               return_value=(9, 16, 1)):
            self.client = RestClientASAr2(**CONNECTION_INFO)

        self.client.ssh_client = mock.MagicMock()
        self.client.connection = mock.MagicMock()
        self.connection = self.client.connection

        self.vserver = CONNECTION_INFO['vserver']
        self.fake_volume = str(uuid.uuid4())
        self.fake_lun = str(uuid.uuid4())

    def _mock_api_error(self, code='fake'):
        return mock.Mock(side_effect=netapp_api.NaApiError(code=code))

    def test_initialization(self):
        """Test ASA r2 client initialization."""
        self.assertIsInstance(self.client, RestClientASAr2)
        self.assertIsInstance(self.client,
                              client_cmode_rest.RestClient)

    def test_init_asar2_features(self):
        """Test ASA r2 specific features initialization."""
        # Test that _init_asar2_features is called during initialization
        with mock.patch.object(RestClientASAr2,
                               '_init_asar2_features') as mock_init:
            with mock.patch.object(RestClientASAr2,
                                   'get_ontap_version',
                                   return_value=(9, 16, 1)):
                RestClientASAr2(**CONNECTION_INFO)

                mock_init.assert_called_once()

    @ddt.data(True, False)
    def test_get_ontapi_version(self, cached):
        """Test that ASA r2 returns (0, 0) for ONTAPI version."""
        result = self.client.get_ontapi_version(cached=cached)
        expected = (0, 0)
        self.assertEqual(expected, result)

    def test_getattr_missing_method(self):
        """Test __getattr__ behavior for missing methods."""
        result = getattr(self.client, 'nonexistent_method', None)
        self.assertIsNone(result)

    def test_send_request_inherits_from_parent(self):
        """Test that send_request inherits behavior from parent class."""
        expected = 'fake_response'
        mock_get_records = self.mock_object(
            self.client, 'get_records',
            mock.Mock(return_value=expected))

        res = self.client.send_request(
            fake_client.FAKE_ACTION_ENDPOINT, 'get',
            body=fake_client.FAKE_BODY,
            query=fake_client.FAKE_HTTP_QUERY, enable_tunneling=False)

        self.assertEqual(expected, res)
        mock_get_records.assert_called_once_with(
            fake_client.FAKE_ACTION_ENDPOINT,
            fake_client.FAKE_HTTP_QUERY, False, 10000)

    def test_send_request_post_inherits_from_parent(self):
        """Test that send_request POST inherits behavior from parent class."""
        expected = (201, 'fake_response')
        mock_invoke = self.mock_object(
            self.client.connection, 'invoke_successfully',
            mock.Mock(return_value=expected))

        res = self.client.send_request(
            fake_client.FAKE_ACTION_ENDPOINT, 'post',
            body=fake_client.FAKE_BODY,
            query=fake_client.FAKE_HTTP_QUERY, enable_tunneling=False)

        self.assertEqual(expected[1], res)
        mock_invoke.assert_called_once_with(
            fake_client.FAKE_ACTION_ENDPOINT, 'post',
            body=fake_client.FAKE_BODY,
            query=fake_client.FAKE_HTTP_QUERY, enable_tunneling=False)

    @ddt.data(
        {'enable_tunneling': True},
        {'enable_tunneling': False}
    )
    @ddt.unpack
    def test_get_records_inherits_from_parent(self, enable_tunneling):
        """Test that get_records inherits behavior from parent class."""
        api_responses = [
            (200, fake_client.VOLUME_GET_ITER_RESPONSE_REST_PAGE),
            (200, fake_client.VOLUME_GET_ITER_RESPONSE_REST_PAGE),
            (200, fake_client.VOLUME_GET_ITER_RESPONSE_REST_LAST_PAGE),
        ]

        self.mock_object(
            self.client.connection, 'invoke_successfully',
            side_effect=copy.deepcopy(api_responses))

        query = {
            'fields': 'name'
        }

        result = self.client.get_records(
            '/storage/volumes/', query=query,
            enable_tunneling=enable_tunneling,
            max_page_length=10)

        num_records = result['num_records']
        self.assertEqual(28, num_records)
        self.assertEqual(28, len(result['records']))

        expected_records = []
        expected_records.extend(api_responses[0][1]['records'])
        expected_records.extend(api_responses[1][1]['records'])
        expected_records.extend(api_responses[2][1]['records'])

        self.assertEqual(expected_records, result['records'])

    def test_send_ems_log_message_inherits_from_parent(self):
        """Test send_ems_log_message inherits behavior"""
        message_dict = {
            'computer-name': '25-dev-vm',
            'event-source': 'Cinder driver NetApp_iSCSI_ASAr2_direct',
            'app-version': 'dummy app version',
            'category': 'provisioning',
            'log-level': '5',
            'auto-support': 'false',
            'event-id': '1',
            'event-description':
                '{"pools": {"vserver": "vserver_name",'
                + '"aggregates": [], "flexvols": ["flexvol_01"]}}'
        }

        body = {
            'computer_name': message_dict['computer-name'],
            'event_source': message_dict['event-source'],
            'app_version': message_dict['app-version'],
            'category': message_dict['category'],
            'severity': 'notice',
            'autosupport_required': message_dict['auto-support'] == 'true',
            'event_id': message_dict['event-id'],
            'event_description': message_dict['event-description'],
        }

        self.mock_object(self.client, '_get_ems_log_destination_vserver',
                         return_value='vserver_name')
        self.mock_object(self.client, 'send_request')

        self.client.send_ems_log_message(message_dict)

        self.client.send_request.assert_called_once_with(
            '/support/ems/application-logs', 'post', body=body)

    def test_inheritance_all_parent_methods_available(self):
        """Test that ASA r2 client has access to all parent methods."""
        # Test that common parent methods are available
        parent_methods = [
            'send_request', 'get_records', 'send_ems_log_message'
        ]

        for method_name in parent_methods:
            self.assertTrue(hasattr(self.client, method_name),
                            f"Method {method_name} should be available")
            self.assertTrue(callable(getattr(self.client, method_name)),
                            f"Method {method_name} should be callable")

    def test_asar2_specific_ontapi_not_supported(self):
        """Test that ASA r2 specifically doesn't support ONTAPI."""
        # This is a key differentiator for ASA r2
        result = self.client.get_ontapi_version()
        self.assertEqual((0, 0), result)

        # No change for cached version
        result_cached = self.client.get_ontapi_version(cached=True)
        self.assertEqual((0, 0), result_cached)

    def test_disaggregated_platform_connection_info(self):
        """Test ASA r2 client works with disaggregated platform settings."""
        # Verify the connection info includes disaggregated flag
        self.assertEqual('True', CONNECTION_INFO['is_disaggregated'])

        # Test that client can be initialized with disaggregated settings
        disaggregated_info = CONNECTION_INFO.copy()
        disaggregated_info['is_disaggregated'] = 'True'

        with mock.patch.object(RestClientASAr2, 'get_ontap_version',
                               return_value=(9, 18, 1)):
            client = RestClientASAr2(**disaggregated_info)
            self.assertIsInstance(client, RestClientASAr2)

    def test_get_cluster_info_success(self):
        """Test successful cluster info retrieval."""
        expected_response = fake_client.GET_CLUSTER_INFO_RESPONSE_REST

        self.mock_object(self.client, 'send_request',
                         return_value=expected_response)

        result = self.client.get_cluster_info()

        expected_query = {'fields': 'name,disaggregated'}
        self.client.send_request.assert_called_once_with(
            '/cluster', 'get', query=expected_query, enable_tunneling=False)
        self.assertEqual(expected_response, result)

    def test_get_cluster_info_exception(self):
        """Test exception handling during cluster info retrieval."""
        self.mock_object(self.client, 'send_request',
                         side_effect=Exception("API error"))

        result = self.client.get_cluster_info()

        expected_query = {'fields': 'name,disaggregated'}
        self.client.send_request.assert_called_once_with(
            '/cluster', 'get', query=expected_query, enable_tunneling=False)
        self.assertIsNone(result)

    def test_get_cluster_info_empty_response(self):
        """Test cluster info retrieval with empty response."""
        self.mock_object(self.client, 'send_request',
                         return_value={})

        result = self.client.get_cluster_info()

        expected_query = {'fields': 'name,disaggregated'}
        self.client.send_request.assert_called_once_with(
            '/cluster', 'get', query=expected_query, enable_tunneling=False)
        self.assertEqual({}, result)

    def test_get_cluster_info_netapp_api_error(self):
        """Test NetApp API error handling during cluster info retrieval."""
        self.mock_object(self.client, 'send_request',
                         side_effect=netapp_api.NaApiError("NetApp API error"))

        result = self.client.get_cluster_info()

        expected_query = {'fields': 'name,disaggregated'}
        self.client.send_request.assert_called_once_with(
            '/cluster', 'get', query=expected_query, enable_tunneling=False)
        self.assertIsNone(result)

    def test_get_cluster_capacity_success(self):
        """Test successful cluster capacity retrieval."""
        expected_response = fake_client.GET_CLUSTER_CAPACITY_RESPONSE_REST

        self.mock_object(self.client, 'send_request',
                         return_value=expected_response)

        result = self.client.get_cluster_capacity()

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get',
            query=expected_query, enable_tunneling=False)

        expected_capacity = {
            'size-total': float(expected_response['block_storage']['size']),
            'size-available':
                float(expected_response['block_storage']['available'])
        }
        self.assertEqual(expected_capacity, result)

    def test_get_cluster_capacity_no_response(self):
        """Test cluster capacity retrieval with no response."""
        self.mock_object(self.client, 'send_request',
                         return_value=None)

        result = self.client.get_cluster_capacity()

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get',
            query=expected_query, enable_tunneling=False)
        self.assertEqual({}, result)

    def test_get_cluster_capacity_missing_block_storage(self):
        """Test cluster capacity retrieval with missing block_storage."""
        response = {'some_other_field': 'value'}

        self.mock_object(self.client, 'send_request',
                         return_value=response)

        result = self.client.get_cluster_capacity()

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get',
            query=expected_query, enable_tunneling=False)

        expected_capacity = {
            'size-total': 0.0,
            'size-available': 0.0
        }
        self.assertEqual(expected_capacity, result)

    def test_get_cluster_capacity_partial_block_storage(self):
        """Test cluster capacity retrieval with partial block_storage."""
        response = {
            'block_storage': {
                'size': 1000000000,
                # missing 'available' field
            }
        }

        self.mock_object(self.client, 'send_request',
                         return_value=response)

        result = self.client.get_cluster_capacity()

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get',
            query=expected_query, enable_tunneling=False)

        expected_capacity = {
            'size-total': 1000000000.0,
            'size-available': 0.0
        }
        self.assertEqual(expected_capacity, result)

    def test_get_cluster_capacity_exception(self):
        """Test exception handling during cluster capacity retrieval."""
        self.mock_object(self.client, 'send_request',
                         side_effect=Exception("API error"))

        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_cluster_capacity)

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get',
            query=expected_query, enable_tunneling=False)

    def test_get_cluster_capacity_netapp_api_error(self):
        """Test NetApp API error handling during cluster capacity retrieval."""
        self.mock_object(self.client, 'send_request',
                         side_effect=netapp_api.NaApiError("NetApp API error"))

        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_cluster_capacity)

        expected_query =\
            {'fields': 'block_storage.size,block_storage.available'}
        self.client.send_request.assert_called_once_with(
            '/storage/cluster', 'get', query=expected_query,
            enable_tunneling=False)

    def test_get_aggregate_disk_types_success(self):
        """Test successful aggregate disk types retrieval."""
        expected_response =\
            fake_client.GET_AGGREGATE_STORAGE_TYPES_RESPONSE_REST

        self.mock_object(self.client, 'send_request',
                         return_value=expected_response)

        result = self.client.get_aggregate_disk_types()

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)
        # Should return array of storage types
        self.assertEqual(['ssd'], result)

    def test_get_aggregate_disk_types_multiple_records(self):
        """Test aggregate disk types retrieval with multiple records."""
        expected_response =\
            fake_client.GET_AGGREGATE_STORAGE_TYPES_MULTIPLE_RESPONSE_REST

        self.mock_object(self.client, 'send_request',
                         return_value=expected_response)

        result = self.client.get_aggregate_disk_types()

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)
        # Should return array with all storage types including duplicates
        self.assertEqual(['ssd', 'ssd'], result)

    def test_get_aggregate_disk_types_empty_records(self):
        """Test aggregate disk types retrieval with empty records."""
        expected_response =\
            fake_client.GET_AGGREGATE_STORAGE_TYPES_EMPTY_RESPONSE_REST

        self.mock_object(self.client, 'send_request',
                         return_value=expected_response)

        result = self.client.get_aggregate_disk_types()

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)
        self.assertIsNone(result)

    def test_get_aggregate_disk_types_missing_block_storage(self):
        """Test aggregate disk types retrieval with missing block_storage."""
        response = {
            "records": [
                {
                    "uuid": "3e5e2865-af43-4d82-a808-8a7222cf0369",
                    "name": "dataFA_2_p0_i1",
                    # missing block_storage field
                }
            ],
            "num_records": 1
        }

        self.mock_object(self.client, 'send_request',
                         return_value=response)

        result = self.client.get_aggregate_disk_types()

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)

        self.assertEqual([], result)

    def test_get_aggregate_disk_types_missing_storage_type(self):
        """Test aggregate disk types retrieval with missing storage_type."""
        response = {
            "records": [
                {
                    "uuid": "3e5e2865-af43-4d82-a808-8a7222cf0369",
                    "name": "dataFA_2_p0_i1",
                    "block_storage": {
                        "primary": {
                            "disk_class": "solid_state",
                            "disk_type": "ssd"
                        }
                        # missing storage_type field
                    }
                }
            ],
            "num_records": 1
        }

        self.mock_object(self.client, 'send_request',
                         return_value=response)

        result = self.client.get_aggregate_disk_types()

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)

        self.assertEqual([], result)

    def test_get_aggregate_disk_types_netapp_api_error(self):
        """Test NetApp API error handling."""
        self.mock_object(self.client, 'send_request',
                         side_effect=netapp_api.NaApiError("NetApp API error"))

        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_aggregate_disk_types)

        expected_query = {'fields': 'name,block_storage.storage_type'}
        self.client.send_request.assert_called_once_with(
            '/storage/aggregates', 'get', query=expected_query,
            enable_tunneling=False)

    def test_get_performance_counter_info_not_supported(self):
        """Performance counter info raises NetAppDriverException."""
        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_performance_counter_info,
                          'system', 'cpu_busy')

    def test_get_performance_instance_uuids_not_supported(self):
        """Performance instance UUIDs raises NetAppDriverException."""
        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_performance_instance_uuids,
                          'system', 'node1')

    def test_get_performance_counters_not_supported(self):
        """Performance counters raises NetAppDriverException."""
        self.assertRaises(netapp_utils.NetAppDriverException,
                          self.client.get_performance_counters,
                          'system', ['uuid1'], ['cpu_busy'])

    def test_create_lun(self):
        metadata = copy.deepcopy(fake_client.LUN_GET_ITER_RESULT[0])
        name = fake.LUN_NAME
        size = 2048
        initial_size = size
        qos_policy_group_is_adaptive = False

        self.mock_object(self.client, '_validate_qos_policy_group')
        self.mock_object(self.client, 'send_request')

        body = {
            'name': name,
            'space.size': str(initial_size),
            'os_type': metadata['OsType'],
            'qos_policy.name': fake.QOS_POLICY_GROUP_NAME
        }

        self.client.create_lun(
            fake.VOLUME_NAME, fake.LUN_NAME, size, metadata,
            qos_policy_group_name=fake.QOS_POLICY_GROUP_NAME,
            qos_policy_group_is_adaptive=qos_policy_group_is_adaptive)

        self.client.send_request.assert_called_once_with(
            '/storage/luns', 'post', body=body)

    @patch('cinder.volume.drivers.netapp.dataontap.client.'
           'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_create_lun_handles_qos_policy(self, mock_send_request):
        mock_send_request.return_value = None

        self.client.create_lun(fake_client.VOLUME_NAME,
                               fake_client.LUN_NAME,
                               1024,
                               {"OsType": "linux"},
                               qos_policy_group_name=(
                                   fake.QOS_POLICY_GROUP_NAME),
                               )

        mock_send_request.assert_called_once_with(
            '/storage/luns', 'post', body={
                'name': fake_client.LUN_NAME.replace("-", "_"),
                'space.size': '1024',
                'os_type': 'linux',
                'qos_policy.name': fake.QOS_POLICY_GROUP_NAME
            }
        )

    @patch('cinder.volume.drivers.netapp.dataontap.client.'
           'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_create_lun_raises_error_on_failure(self, mock_send_request):
        mock_send_request.side_effect = netapp_api.NaApiError
        self.assertRaises(
            netapp_api.NaApiError,
            self.client.create_lun,
            fake.VOLUME_NAME, fake.LUN_NAME, 1024, {"OsType": "linux"}
        )

    @patch('cinder.volume.drivers.netapp.dataontap.client.'
           'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_destroy_lun(self, mock_send_request):
        mock_send_request.return_value = None

        self.client.destroy_lun(fake.LUN_PATH, force=True)
        lun_name = self.client._get_backend_lun_or_namespace(
            fake.LUN_PATH
        )

        mock_send_request.assert_called_once_with(
            '/storage/luns/', 'delete', query={
                'name': lun_name,
                'allow_delete_while_mapped': 'true',
            }
        )

    @patch('cinder.volume.drivers.netapp.dataontap.client.'
           'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_destroy_lun_handles_non_forced_deletion(self, mock_send_request):
        mock_send_request.return_value = None

        self.client.destroy_lun(fake.LUN_PATH, force=False)
        lun_name = self.client._get_backend_lun_or_namespace(
            fake.LUN_PATH
        )

        mock_send_request.assert_called_once_with(
            '/storage/luns/', 'delete', query={
                'name': lun_name,
            }
        )

    @patch('cinder.volume.drivers.netapp.dataontap.client.'
           'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_destroy_lun_raises_error_on_failure(self, mock_send_request):
        mock_send_request.side_effect = netapp_api.NaApiError
        self.assertRaises(
            netapp_api.NaApiError,
            self.client.destroy_lun,
            fake.LUN_PATH, force=True,
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_create_namespace(self, mock_send_request):
        self.client.create_namespace(fake_client.VOLUME_NAME,
                                     fake_client.NAMESPACE_NAME,
                                     2048, {'OsType': 'linux'}
                                     )

        mock_send_request.assert_called_once_with(
            '/storage/namespaces', 'post', body={
                'name': fake_client.NAMESPACE_NAME,
                'space.size': '2048',
                'os_type': 'linux'
            }
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_destroy_namespace(self, mock_send_request):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.NAMESPACE_NAME)
        self.client.destroy_namespace(fake.PATH_NAMESPACE, force=True)
        mock_send_request.assert_called_once_with(
            '/storage/namespaces', 'delete', query={
                'name': fake.NAMESPACE_NAME,
                'svm': self.client.vserver,
                'allow_delete_while_mapped': 'true'
            }
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.map_lun')
    def test_map_lun(self, mock_super_map_lun):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.LUN_NAME
        )
        mock_super_map_lun.return_value = 'result'

        result = self.client.map_lun(fake.LUN_PATH, 'igroup1', 42)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.LUN_PATH
        )
        mock_super_map_lun.assert_called_once_with(fake.LUN_NAME,
                                                   'igroup1',
                                                   42)
        self.assertEqual(result, 'result')

    @mock.patch('cinder.volume.drivers.netapp.dataontap.'
                'client.client_cmode_rest.RestClient.get_lun_map')
    def test_get_lun_map(self, mock_super_get_lun_map):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.LUN_NAME)
        mock_super_get_lun_map.return_value = [
            {'initiator-group': 'igroup1', 'lun-id': 1, 'vserver': 'svm1'}
        ]

        result = self.client.get_lun_map(fake.LUN_NAME)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.LUN_NAME
        )
        mock_super_get_lun_map.assert_called_once_with(fake.LUN_NAME)
        self.assertEqual(result, [
            {'initiator-group': 'igroup1', 'lun-id': 1, 'vserver': 'svm1'}
        ])

    @mock.patch('cinder.volume.drivers.netapp.dataontap.'
                'client.client_cmode_rest.RestClient.unmap_lun')
    def test_unmap_lun(self, mock_super_unmap_lun):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.LUN_NAME
        )
        self.client.unmap_lun(fake.LUN_NAME, fake.IGROUP1, )
        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.LUN_NAME
        )
        mock_super_unmap_lun.assert_called_once_with(
            fake.LUN_NAME, fake.IGROUP1
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.get_lun_by_args')
    def test_get_lun_by_args_with_path(self, mock_super_get_lun_by_args):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.LUN_NAME
        )
        mock_super_get_lun_by_args.return_value = 'result'

        path_arg = {'path': fake.LUN_PATH}
        result = self.client.get_lun_by_args(path=path_arg)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            path_arg
        )
        self.assertEqual(path_arg['path'], fake.LUN_NAME)
        mock_super_get_lun_by_args.assert_called_once_with(path=path_arg)
        self.assertEqual(result, 'result')

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.get_lun_by_args')
    def test_get_lun_by_args_without_path(self, mock_super_get_lun_by_args):
        mock_super_get_lun_by_args.return_value = 'result'
        result = self.client.get_lun_by_args(path=None)
        mock_super_get_lun_by_args.assert_called_once_with(path=None)
        self.assertEqual(result, 'result')

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.map_namespace')
    def test_maps_namespace(self, mock_super_map_namespace):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.NAMESPACE_NAME
        )
        mock_super_map_namespace.return_value = 'namespace-uuid'

        result = self.client.map_namespace(fake.PATH_NAMESPACE, fake.SUBSYSTEM)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.PATH_NAMESPACE
        )
        mock_super_map_namespace.assert_called_once_with(
            fake.NAMESPACE_NAME, fake.SUBSYSTEM
        )
        self.assertEqual(result, 'namespace-uuid')

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.map_namespace')
    def test_maps_namespace_with_path_containing_hyphens(
            self, mock_super_map_namespace):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value='name_space_2')
        mock_super_map_namespace.return_value = 'uuid-2'

        result = self.client.map_namespace('/vol/vol1/name-space-2',
                                           'subsystem2')

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            '/vol/vol1/name-space-2')
        mock_super_map_namespace.assert_called_once_with(
            'name_space_2', 'subsystem2')
        self.assertEqual(result, 'uuid-2')

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.unmap_namespace')
    def test_unmaps_namespace_with_valid_path_and_subsystem(
            self, mock_super_unmap_namespace):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.NAMESPACE_NAME)

        self.client.unmap_namespace(fake.PATH_NAMESPACE, fake.SUBSYSTEM)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.PATH_NAMESPACE
        )
        mock_super_unmap_namespace.assert_called_once_with(
            fake.NAMESPACE_NAME, fake.SUBSYSTEM
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.unmap_namespace')
    def test_unmaps_namespace_with_path_containing_special_characters(
            self, mock_super_unmap_namespace):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value='namespace_special')

        self.client.unmap_namespace('/vol/vol1/namespace-special',
                                    'subsystem2')

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            '/vol/vol1/namespace-special')
        mock_super_unmap_namespace.assert_called_once_with(
            'namespace_special',
            'subsystem2'
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.get_namespace_map')
    def test_get_namespace_map(self, mock_super_get_namespace_map):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.NAMESPACE_NAME
        )
        mock_super_get_namespace_map.return_value = {
            'namespace_map': 'details'}

        result = self.client.get_namespace_map(fake.PATH_NAMESPACE)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.PATH_NAMESPACE
        )
        mock_super_get_namespace_map.assert_called_once_with(
            fake.NAMESPACE_NAME)
        self.assertEqual(result, {'namespace_map': 'details'})

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.get_namespace_map')
    def test_get_namespace_map_with_path_containing_special_characters(
            self, mock_super_get_namespace_map):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value='namespace_special')
        mock_super_get_namespace_map.return_value = {
            'namespace_map': 'special_details'}

        result = self.client.get_namespace_map('/vol/vol1/namespace-special')

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            '/vol/vol1/namespace-special')
        mock_super_get_namespace_map.assert_called_once_with(
            'namespace_special')
        self.assertEqual(result, {
            'namespace_map': 'special_details'
        })

    @mock.patch('cinder.volume.drivers.netapp.dataontap.client.'
                'client_cmode_rest.RestClient.get_namespace_map')
    def test_returns_none_when_namespace_map_not_found(
            self, mock_super_get_namespace_map):
        mock_super_get_namespace_map.return_value = None
        result = self.client.get_namespace_map('/vol/vol1/namespace3')

        mock_super_get_namespace_map.assert_called_once_with('namespace3')
        self.assertIsNone(result)

    @mock.patch(
        'cinder.volume.drivers.netapp.dataontap.client.'
        'client_cmode_rest_asar2.RestClientASAr2._lun_update_by_path')
    def test_resizes_lun(self, mock_lun_update_by_path):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.LUN_NAME)

        self.client.do_direct_resize(fake.LUN_PATH, 10)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.LUN_PATH)
        mock_lun_update_by_path.assert_called_once_with(
            fake.LUN_NAME, {'name': fake.LUN_NAME, 'space.size': 10})

    @mock.patch(
        'cinder.volume.drivers.netapp.dataontap.client.'
        'client_cmode_rest_asar2.RestClientASAr2._lun_update_by_path'
    )
    def test_resize_lun_with_invalid_path(self, mock_lun_update_by_path):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=None)
        self.client.do_direct_resize('/vol/vol1/invalid_lun', 53)
        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            '/vol/vol1/invalid_lun')
        mock_lun_update_by_path.assert_not_called()

    @mock.patch('cinder.volume.drivers.netapp.dataontap.'
                'client.client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_resizes_namespace(self, mock_send_request):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=fake.NAMESPACE_NAME)

        self.client.namespace_resize(fake.PATH_NAMESPACE, fake.SIZE)

        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            fake.PATH_NAMESPACE)
        mock_send_request.assert_called_once_with(
            '/storage/namespaces',
            'patch',
            body={'space.size': fake.SIZE},
            query={'name': fake.NAMESPACE_NAME}
        )

    @mock.patch('cinder.volume.drivers.netapp.dataontap.'
                'client.client_cmode_rest_asar2.RestClientASAr2.send_request')
    def test_resize_namespace_with_invalid_path(self, mock_send_request):
        self.client._get_backend_lun_or_namespace = mock.Mock(
            return_value=None)
        self.client.namespace_resize('/vol/vol1/invalid_namespace', 5368)
        self.client._get_backend_lun_or_namespace.assert_called_once_with(
            '/vol/vol1/invalid_namespace')
        mock_send_request.assert_not_called()



#    Copyright (c) 2023 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import units as o_units

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom
from cinder.volume.drivers.open_e.jovian_common import rest

LOG = logging.getLogger(__name__)


class JovianDSSDriver(object):

    def __init__(self, config):

        self.configuration = config
        self._pool = self.configuration.get('jovian_pool', 'Pool-0')
        self.jovian_iscsi_target_portal_port = self.configuration.get(
            'target_port', 3260)

        self.jovian_target_prefix = self.configuration.get(
            'target_prefix',
            'iqn.2020-04.com.open-e.cinder:')
        self.jovian_chap_pass_len = self.configuration.get(
            'chap_password_len', 12)
        self.block_size = (
            self.configuration.get('jovian_block_size', '64K'))
        self.jovian_sparse = (
            self.configuration.get('san_thin_provision', True))
        self.jovian_ignore_tpath = self.configuration.get(
            'jovian_ignore_tpath', None)
        self.jovian_hosts = self.configuration.get(
            'san_hosts', [])

        self.ra = rest.JovianRESTAPI(config)

    def rest_config_is_ok(self):
        """Check config correctness by checking pool availability"""

        return self.ra.is_pool_exists()

    def get_active_ifaces(self):
        """Return list of ip addresses for iSCSI connection"""

        return self.jovian_hosts

    def get_provider_location(self, volume_name):
        """Return volume iscsiadm-formatted provider location string."""
        return '%(host)s:%(port)s,1 %(name)s 0' % {
            'host': self.ra.get_active_host(),
            'port': self.jovian_iscsi_target_portal_port,
            'name': self._get_target_name(volume_name)}

    def create_volume(self, volume_id, volume_size, sparse=False,
                      block_size=None):
        """Create a volume.

        :param str volume_id: volume id
        :param int volume_size: size in Gi
        :param bool sparse: thin or thick volume flag (default thin)
        :param int block_size: size of block (default None)

        :return: None
        """
        vname = jcom.vname(volume_id)
        LOG.debug("Create volume:%(name)s with size:%(size)s",
                  {'name': volume_id, 'size': volume_size})

        self.ra.create_lun(vname,
                           volume_size * o_units.Gi,
                           sparse=sparse,
                           block_size=block_size)
        return

    def _promote_newest_delete(self, vname, snapshots=None):
        '''Promotes and delete volume

        This function deletes volume.
        It will promote volume if needed before deletion.

        :param str vname: physical volume id
        :param list snapshots: snapshot data list (default None)

        :return: None
        '''

        if snapshots is None:
            try:
                snapshots = self.ra.get_snapshots(vname)
            except jexc.JDSSResourceNotFoundException:
                LOG.debug('volume %s do not exists, it was already '
                          'deleted', vname)
                return

        bsnaps = self._list_busy_snapshots(vname, snapshots)

        if len(bsnaps) != 0:

            promote_target = None

            sname = jcom.get_newest_snapshot_name(bsnaps)

            for snap in bsnaps:
                if snap['name'] == sname:
                    cvnames = jcom.snapshot_clones(snap)
                    for cvname in cvnames:
                        if jcom.is_volume(cvname):
                            promote_target = cvname
                        if jcom.is_snapshot(cvname):
                            self._promote_newest_delete(cvname)
                        if jcom.is_hidden(cvname):
                            self._promote_newest_delete(cvname)
                    break

            if promote_target is None:
                self._promote_newest_delete(vname)
                return

            self.ra.promote(vname, sname, promote_target)

        self._delete_vol_with_source_snap(vname, recursive=True)

    def _delete_vol_with_source_snap(self, vname, recursive=False):
        '''Delete volume and its source snapshot if required

        This function deletes volume.
        If volume is a clone it will check its source snapshot if
        one is originates from volume to delete.

        :param str vname: physical volume id
        :param bool recursive: recursive flag (default False)

        :return: None
        '''
        vol = None

        try:
            vol = self.ra.get_lun(vname)
        except jexc.JDSSResourceNotFoundException:
            LOG.debug('unable to get volume %s info, '
                      'assume it was already deleted', vname)
            return
        try:
            self.ra.delete_lun(vname,
                               force_umount=True,
                               recursively_children=recursive)
        except jexc.JDSSResourceNotFoundException:
            LOG.debug('volume %s do not exists, it was already '
                      'deleted', vname)
            return

        if vol is not None and \
                'origin' in vol and \
                vol['origin'] is not None:
            if jcom.is_volume(jcom.origin_snapshot(vol)) or \
                    jcom.is_hidden(jcom.origin_snapshot(vol)) or \
                    (jcom.vid_from_sname(jcom.origin_snapshot(vol)) ==
                     jcom.idname(vname)):
                self.ra.delete_snapshot(jcom.origin_volume(vol),
                                        jcom.origin_snapshot(vol),
                                        recursively_children=True,
                                        force_umount=True)

    def _clean_garbage_resources(self, vname, snapshots=None):
        '''Removes resources that is not related to volume

        Goes through volume snapshots and it clones to identify one
        that is clearly not related to vname volume and therefore
        have to be deleted.

        :param str vname: physical volume id
        :param list snapshots: list of snapshot info dictionaries

        :return: updated list of snapshots
        '''

        if snapshots is None:
            try:
                snapshots = self.ra.get_snapshots(vname)
            except jexc.JDSSResourceNotFoundException:
                LOG.debug('volume %s do not exists, it was already '
                          'deleted', vname)
                return
        update = False
        for snap in snapshots:
            if jcom.is_volume(jcom.sname_from_snap(snap)):
                cvnames = jcom.snapshot_clones(snap)
                if len(cvnames) == 0:
                    self._delete_snapshot(vname, jcom.sname_from_snap(snap))
                    update = True
            if jcom.is_snapshot(jcom.sname_from_snap(snap)):
                cvnames = jcom.snapshot_clones(snap)
                for cvname in cvnames:
                    if jcom.is_hidden(cvname):
                        self._promote_newest_delete(cvname)
                        update = True
                    if jcom.is_snapshot(cvname):
                        if jcom.idname(vname) != jcom.vid_from_sname(cvname):
                            self._promote_newest_delete(cvname)
                            update = True
        if update:
            snapshots = self.ra.get_snapshots(vname)
        return snapshots

    def _list_busy_snapshots(self, vname, snapshots,
                             exclude_dedicated_volumes=False) -> list:
        """List all volume snapshots with clones

        Goes through provided list of snapshots.
        If additional parameters are given, will filter list of snapshots
        accordingly.

        Keyword arguments:
        :param str vname: zvol id
        :param list snapshots: list of snapshots data dicts
        :param bool exclude_dedicated_volumes: list snapshots that has clones
                                        (default False)

        :return: filtered list of snapshot data dicts
        :rtype: list
        """

        out = []
        for snap in snapshots:
            clones = jcom.snapshot_clones(snap)
            add = False
            for cvname in clones:
                if exclude_dedicated_volumes and jcom.is_volume(cvname):
                    continue
                add = True
            if add:
                out.append(snap)

        return out

    def _clean_volume_snapshots_mount_points(self, vname, snapshots):
        update = False
        for snap in snapshots:
            clones = jcom.snapshot_clones(snap)
            for cname in [c for c in clones if jcom.is_snapshot(c)]:
                update = True
                self._delete_volume(cname, cascade=True)
        if update:
            snapshots = self.ra.get_snapshots(vname)
        return snapshots

    def _delete_volume(self, vname, cascade=False):
        """_delete_volume delete routine containing delete logic

        :param str vname: physical volume id
        :param bool cascade: flag for cascade volume deletion
            with its snapshots

        :return: None
        """
        try:
            self.ra.delete_lun(vname,
                               force_umount=True,
                               recursively_children=cascade)
        except jexc.JDSSResourceIsBusyException:
            LOG.debug('unable to conduct direct volume %s deletion', vname)
        except jexc.JDSSResourceNotFoundException:
            LOG.debug('volume %s do not exists, it was already '
                      'deleted', vname)
            return
        except jexc.JDSSRESTException as jerr:
            LOG.debug(
                "Unable to delete physical volume %(volume)s "
                "with error %(err)s.", {
                    "volume": vname,
                    "err": jerr})
        else:
            LOG.debug('in place deletion suceeded')
            return

        snapshots = None
        try:
            snapshots = self.ra.get_snapshots(vname)
        except jexc.JDSSResourceNotFoundException:
            LOG.debug('volume %s do not exists, it was already '
                      'deleted', vname)
            return

        if cascade is False:
            bsnaps = self._list_busy_snapshots(vname,
                                               snapshots,
                                               exclude_dedicated_volumes=True)
            if len(bsnaps) > 0:
                raise exception.VolumeIsBusy('Volume has snapshots')

        snaps = self._clean_garbage_resources(vname, snapshots)
        snaps = self._clean_volume_snapshots_mount_points(vname, snapshots)

        self._promote_newest_delete(vname, snapshots=snaps)

    def delete_volume(self, volume_name, cascade=False):
        """Delete volume

        :param volume: volume reference
        :param cascade: remove snapshots of a volume as well
        """
        vname = jcom.vname(volume_name)

        LOG.debug('deleting volume %s', vname)

        self._delete_volume(vname, cascade=cascade)

    def _clone_object(self, cvname, sname, ovname,
                      sparse=None,
                      create_snapshot=False):
        """Creates a clone of specified object

        Will create snapshot if it is not provided

        :param str cvname: clone volume name
        :param str sname: snapshot name
        :param str ovname: original volume name
        :param bool sparse: sparse property of new volume
        :param bool create_snapshot:
        """
        LOG.debug('cloning %(ovname)s to %(coname)s', {
            "ovname": ovname,
            "coname": cvname})

        if create_snapshot:
            self.ra.create_snapshot(ovname, sname)
        try:
            self.ra.create_volume_from_snapshot(
                cvname,
                sname,
                ovname,
                sparse=sparse)
        except jexc.JDSSException as jerr:
            # This is a garbage collecting section responsible for cleaning
            # all the mess of request failed
            if create_snapshot:
                try:
                    self.ra.delete_snapshot(ovname,
                                            cvname,
                                            recursively_children=True,
                                            force_umount=True)
                except jexc.JDSSException as jerrd:
                    LOG.warning("Because of %s physical snapshot %s of volume"
                                " %s have to be removed manually",
                                jerrd,
                                sname,
                                ovname)

            raise jerr

    def resize_volume(self, volume_name, new_size):
        """Extend an existing volume.

        :param str volume_name: volume id
        :param int new_size: volume new size in Gi
        """
        LOG.debug("Extend volume:%(name)s to size:%(size)s",
                  {'name': volume_name, 'size': new_size})

        self.ra.extend_lun(jcom.vname(volume_name),
                           int(new_size) * o_units.Gi)

    def create_cloned_volume(self,
                             clone_name,
                             volume_name,
                             size,
                             snapshot_name=None,
                             sparse=False):
        """Create a clone of the specified volume.

        :param str clone_name: new volume id
        :param volume_name: original volume id
        :param int size: size in Gi
        :param str snapshot_name: openstack snapshot id to use for cloning
        :param bool sparse: sparse flag
        """
        cvname = jcom.vname(clone_name)

        ovname = jcom.vname(volume_name)

        LOG.debug('clone volume %(id)s to %(id_clone)s', {
            "id": volume_name,
            "id_clone": clone_name})

        if snapshot_name:
            sname = jcom.sname(snapshot_name, volume_name)
            self._clone_object(cvname, sname, ovname,
                               create_snapshot=False,
                               sparse=sparse)
        else:
            sname = jcom.vname(clone_name)
            self._clone_object(cvname, sname, ovname,
                               create_snapshot=True,
                               sparse=sparse)

        clone_size = 0

        try:
            clone_size = int(self.ra.get_lun(cvname)['volsize'])
        except jexc.JDSSException as jerr:

            self.delete_volume(clone_name, cascade=False)
            raise exception.VolumeBackendAPIException(
                _("Fail in cloning volume %(vol)s to %(clone)s.") % {
                    'vol': volume_name, 'clone': clone_name}) from jerr

        try:
            if int(clone_size) < o_units.Gi * int(size):
                self.resize_volume(clone_name, int(size))

        except jexc.JDSSException as jerr:
            # If volume can't be set to a proper size make sure to clean it
            # before failing
            try:
                self.delete_volume(clone_name, cascade=False)
            except jexc.JDSSException as jerrex:
                LOG.warning("Error %s during cleaning failed volume %s",
                            jerrex, volume_name)
                raise jerr from jerrex

    def create_snapshot(self, snapshot_name, volume_name):
        """Create snapshot of existing volume.

        :param str snapshot_name: new snapshot id
        :param str volume_name: original volume id
        """
        LOG.debug('create snapshot %(snap)s for volume %(vol)s', {
            'snap': snapshot_name,
            'vol': volume_name})

        vname = jcom.vname(volume_name)
        sname = jcom.sname(snapshot_name, volume_name)

        self.ra.create_snapshot(vname, sname)

    def create_export_snapshot(self, snapshot_name, volume_name,
                               provider_auth):
        """Creates iscsi resources needed to start using snapshot

        :param str snapshot_name: openstack snapshot id
        :param str volume_name: openstack volume id
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        """

        sname = jcom.sname(snapshot_name, volume_name)
        ovname = jcom.vname(volume_name)
        self._clone_object(sname, sname, ovname,
                           sparse=True,
                           create_snapshot=False)
        try:
            self._ensure_target_volume(snapshot_name, sname, provider_auth,
                                       ro=True)
        except jexc.JDSSException as jerr:
            self._delete_volume(sname, cascade=True)
            raise jerr

    def remove_export(self, volume_name):
        """Remove iscsi target created to make volume attachable

        :param str volume_name: openstack volume id
        """
        vname = jcom.vname(volume_name)
        try:
            self._remove_target_volume(volume_name, vname)
        except jexc.JDSSException as jerr:
            LOG.warning(jerr)

    def remove_export_snapshot(self, snapshot_name, volume_name):
        """Remove tmp vol and iscsi target created to make snap attachable

        :param str snapshot_name: openstack snapshot id
        :param str volume_name: openstack volume id
        """

        sname = jcom.sname(snapshot_name, volume_name)

        try:
            self._remove_target_volume(snapshot_name, sname)
        except jexc.JDSSException as jerr:
            self._delete_volume(sname, cascade=True)
            raise jerr

        self._delete_volume(sname, cascade=True)

    def _delete_snapshot(self, vname, sname):
        """Delete snapshot

        This method will delete snapshot mount point and snapshot if possible

        :param str vname: zvol name
        :param dict snap: snapshot info dictionary

        :return: None
        """

        try:
            self.ra.delete_snapshot(vname, sname, force_umount=True)
        except jexc.JDSSResourceIsBusyException:
            LOG.debug('Direct deletion of snapshot %s failed', vname)
        else:
            return

        snap = self.ra.get_snapshot(vname, sname)

        clones = jcom.snapshot_clones(snap)
        busy = False
        for cvname in clones:
            if jcom.is_snapshot(cvname):
                self._promote_newest_delete(cvname)
            if jcom.is_volume(cvname):
                LOG.debug('Will not delete snap %(snap)s,'
                          'becasue it is used by %(vol)s',
                          {'snap': sname,
                           'vol': cvname})
                busy = True
        if busy:
            return
        try:
            self.ra.delete_snapshot(vname, sname, force_umount=True)
        except jexc.JDSSResourceIsBusyException:
            LOG.debug('Unable to delete snap %(snap)s because it is busy',
                      {'snap': jcom.sname_from_snap(snap)})

    def delete_snapshot(self, volume_name, snapshot_name):
        """Delete snapshot of existing volume.

        :param str volume_name: volume id
        :param str snapshot_name: snapshot id
        """
        vname = jcom.vname(volume_name)
        sname = jcom.sname(snapshot_name, volume_name)

        self._delete_snapshot(vname, sname)

    def _ensure_target_volume(self, id, vid, provider_auth, ro=False):
        """Checks if target configured properly and volume is attached to it

        :param str id: id that would be used for target naming
        :param str vname: physical volume id
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        """
        LOG.debug("ensure volume %s assigned to a proper target", id)

        target_name = self._get_target_name(id)

        if not provider_auth:
            msg = _("volume %s is missing provider_auth") % jcom.idname(id)
            raise jexc.JDSSException(msg)

        if not self.ra.is_target(target_name):

            return self._create_target_volume(id, vid, provider_auth)

        if not self.ra.is_target_lun(target_name, vid):
            self._attach_target_volume(target_name, vid)

        (__, auth_username, auth_secret) = provider_auth.split()
        chap_cred = {"name": auth_username,
                     "password": auth_secret}

        try:
            users = self.ra.get_target_user(target_name)
            if len(users) == 1:
                if users[0]['name'] == chap_cred['name']:
                    return
                self.ra.delete_target_user(
                    target_name,
                    users[0]['name'])
            for user in users:
                self.ra.delete_target_user(
                    target_name,
                    user['name'])
            self._set_target_credentials(target_name, chap_cred)

        except jexc.JDSSException as jerr:
            self.ra.delete_target(target_name)
            raise exception.VolumeBackendAPIException(jerr)

    def _get_target_name(self, volume_id):
        """Return iSCSI target name to access volume."""
        return f'{self.jovian_target_prefix}{volume_id}'

    def _get_iscsi_properties(self, volume_id, provider_auth, multipath=False):
        """Return dict according to cinder/driver.py implementation.

        :param volume_id: UUID of volume, might take snapshot UUID
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        :return:
        """
        tname = self._get_target_name(volume_id)
        iface_info = []
        if multipath:
            iface_info = self.get_active_ifaces()
            if not iface_info:
                raise exception.InvalidConfigurationValue(
                    _('No available interfaces '
                      'or config excludes them'))

        iscsi_properties = {}

        if multipath:
            iscsi_properties['target_iqns'] = []
            iscsi_properties['target_portals'] = []
            iscsi_properties['target_luns'] = []
            LOG.debug('tpaths %s.', iface_info)
            for iface in iface_info:
                iscsi_properties['target_iqns'].append(
                    self._get_target_name(volume_id))
                iscsi_properties['target_portals'].append(
                    iface +
                    ":" +
                    str(self.jovian_iscsi_target_portal_port))
                iscsi_properties['target_luns'].append(0)
        else:
            iscsi_properties['target_iqn'] = tname
            iscsi_properties['target_portal'] = (
                self.ra.get_active_host() +
                ":" +
                str(self.jovian_iscsi_target_portal_port))

        iscsi_properties['target_discovered'] = False

        if provider_auth:
            (auth_method, auth_username, auth_secret) = provider_auth.split()

            iscsi_properties['auth_method'] = auth_method
            iscsi_properties['auth_username'] = auth_username
            iscsi_properties['auth_password'] = auth_secret

        iscsi_properties['target_lun'] = 0
        return iscsi_properties

    def _remove_target_volume(self, id, vid):
        """_remove_target_volume

        Ensure that volume is not attached to target and target do not exists.
        """

        target_name = self._get_target_name(id)
        LOG.debug("remove export")
        LOG.debug("detach volume:%(vol)s from target:%(targ)s.", {
            'vol': id,
            'targ': target_name})

        try:
            self.ra.detach_target_vol(target_name, vid)
        except jexc.JDSSResourceNotFoundException as jerrrnf:
            LOG.debug('failed to remove resource %(t)s because of %(err)s', {
                't': target_name,
                'err': jerrrnf.args[0]})
        except jexc.JDSSException as jerr:
            LOG.warning('failed to Terminate_connection for target %(targ)s '
                        'because of: %(err)s', {'targ': target_name,
                                                'err': jerr.args[0]})
            raise jerr

        LOG.debug("delete target: %s", target_name)

        try:
            self.ra.delete_target(target_name)
        except jexc.JDSSResourceNotFoundException as jerrrnf:
            LOG.debug('failed to remove resource %(target)s because '
                      'of %(err)s',
                      {'target': target_name, 'err': jerrrnf.args[0]})

        except jexc.JDSSException as jerr:
            LOG.warning('Failed to Terminate_connection for target %(targ)s '
                        'because of: %(err)s ',
                        {'targ': target_name, 'err': jerr.args[0]})

            raise jerr

    def ensure_export(self, volume_id, provider_auth):

        vname = jcom.vname(volume_id)

        self._ensure_target_volume(volume_id, vname, provider_auth)

    def initialize_connection(self, volume_id, provider_auth,
                              snapshot_id=None,
                              multipath=False):
        """Ensures volume is ready for connection and return connection data

        Ensures that particular volume is ready to be used over iscsi
        with credentials provided in provider_auth
        If snapshot name is provided method will ensure that connection
        leads to read only volume object associated with particular snapshot

        :param str volume_id: Volume id string
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        :param str snapshot_id: id of snapshot that should be connected
        :param bool multipath: specifies if multipath should be used
        """

        id_of_disk_to_attach = volume_id
        vid = jcom.vname(volume_id)
        if provider_auth is None:
            raise jexc.JDSSException(_("CHAP credentials missing"))
        if snapshot_id:
            id_of_disk_to_attach = snapshot_id
            vid = jcom.sname(snapshot_id, volume_id)
        iscsi_properties = self._get_iscsi_properties(id_of_disk_to_attach,
                                                      provider_auth,
                                                      multipath=multipath)
        if snapshot_id:
            self._ensure_target_volume(id_of_disk_to_attach,
                                       vid,
                                       provider_auth,
                                       mode='ro')
        else:
            self._ensure_target_volume(id_of_disk_to_attach,
                                       vid,
                                       provider_auth)

        LOG.debug(
            "initialize_connection for physical disk %(vid)s with %(id)s",
            {'vid': vid, 'id': id_of_disk_to_attach})

        return {
            'driver_volume_type': 'iscsi',
            'data': iscsi_properties,
        }

    def _create_target_volume(self, id, vid, provider_auth):
        """Creates target and attach volume to it

        :param id: uuid of particular resource
        :param vid: physical volume id, might identify snapshot mount
        :param str provider_auth: space-separated triple
              '<auth method> <auth username> <auth password>'
        :return:
        """
        LOG.debug("create target and attach volume %s to it", vid)

        target_name = self._get_target_name(id)

        (__, auth_username, auth_secret) = provider_auth.split()
        chap_cred = {"name": auth_username,
                     "password": auth_secret}

        # Create target
        self.ra.create_target(target_name, use_chap=True)

        # Attach volume
        self._attach_target_volume(target_name, vid)

        # Set credentials
        self._set_target_credentials(target_name, chap_cred)

    def _attach_target_volume(self, target_name, vname):
        """Attach target to volume and handles exceptions

        Attempts to set attach volume to specific target.
        In case of failure will remove target.
        :param target_name: name of target
        :param use_chap: flag for using chap
        """
        try:
            self.ra.attach_target_vol(target_name, vname)
        except jexc.JDSSException as jerr:
            msg = ('Unable to attach volume {volume} to target {target} '
                   'because of {error}.')
            LOG.warning(msg, {"volume": vname,
                              "target": target_name,
                              "error": jerr})
            self.ra.delete_target(target_name)
            raise jerr

    def _set_target_credentials(self, target_name, cred):
        """Set CHAP configuration for target and handle exceptions

        Attempts to set CHAP credentials for specific target.
        In case of failure will remove target.
        :param target_name: name of target
        :param cred: CHAP user name and password
        """
        try:
            self.ra.create_target_user(target_name, cred)

        except jexc.JDSSException as jerr:
            try:
                self.ra.delete_target(target_name)
            except jexc.JDSSException:
                pass

            err_msg = (('Unable to create user %(user)s '
                        'for target %(target)s '
                        'because of %(error)s.') % {
                            'target': target_name,
                            'user': cred['name'],
                            'error': jerr})

            LOG.error(err_msg)
            raise jexc.JDSSException(_(err_msg))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations. For each database backend supported by cinder,
the test case runs a series of test cases to ensure that migrations work
properly and that no data loss occurs if possible.
"""

import functools
from unittest import mock

from alembic import command as alembic_api
from alembic import script as alembic_script
import fixtures
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils as db_utils
from oslo_log.fixture import logging_error as log_fixture
from oslotest import base as test_base
import sqlalchemy

from cinder.db import migration
from cinder.db.sqlalchemy import api
from cinder.db.sqlalchemy import models
from cinder.tests import fixtures as cinder_fixtures


def prevent_drop_alter(func):
    """Decorator to prevent dropping or altering tables and columns.

    With rolling upgrades we shouldn't blindly allow dropping or altering
    tables and columns, since that can easily break them.

    Dropping and altering should be done in a backward-compatible manner.  A
    more detailed explanation is provided in Cinder's developer documentation.

    To properly work, the first parameter of the decorated method must be a
    class or instance with the DROP_ALTER_EXCEPTIONS and FORBIDDEN_METHODS
    attribute, and the second parameter must be the version (legacy migrations)
    or revision (alembic migrations).

    Reviewers should be very careful when adding exceptions to
    DROP_ALTER_EXCEPTIONS and make sure that in the previous release there was
    nothing using that column, not even an ORM model (unless the whole ORM
    model was not being used)
    """

    @functools.wraps(func)
    def wrapper(self, revision, *args, **kwargs):
        exceptions = getattr(self, 'DROP_ALTER_EXCEPTIONS', [])
        do_ban = revision not in exceptions

        patchers = []

        if do_ban:
            forbidden = getattr(self, 'FORBIDDEN_METHODS', [])
            for path in forbidden:
                txt = (f'Migration {revision}: Operation {path}() is not '
                       'allowed in a DB migration')
                patcher = mock.patch(path, autospec=True,
                                     side_effect=Exception(txt))
                patcher.start()
                patchers.append(patcher)

        try:
            return func(self, revision, *args, **kwargs)
        finally:
            for patcher in patchers:
                patcher.stop()

    return wrapper


class CinderModelsMigrationsSync(test_migrations.ModelsMigrationsSync):
    """Test sqlalchemy-migrate migrations."""

    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4

    def setUp(self):
        # Ensure BaseTestCase's ConfigureLogging fixture is disabled since
        # we're using our own (StandardLogging).
        with fixtures.EnvironmentVariable('OS_LOG_CAPTURE', '0'):
            super().setUp()

        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.useFixture(cinder_fixtures.WarningsFixture())
        self.useFixture(cinder_fixtures.StandardLogging())

        self.engine = enginefacade.writer.get_engine()
        self.patch(api, 'get_engine', self.get_engine)

    def db_sync(self, engine):
        migration.db_sync(engine=self.engine)

    def get_engine(self):
        return self.engine

    def get_metadata(self):
        return models.BASE.metadata

    def filter_metadata_diff(self, diff):
        """Filter out allowed differences between DB ORM model and actual DB

        We want to keep the DB ORM (models.py) and DB tables in sync, so the
        test_models_sync test checks for discrepancies between them.

        Due to the rolling upgrades feature there are cases where we will have
        the model and DB out of sync for 1 release, as we stop using it in the
        ORM first and then remove it from the DB in the next, so we use this
        method to allow such discrepancies.  We must add a TODO item on the
        filtering code with the release the filtering must be removed.

        The diff parameter is a list of diff directives, which can be a tuple
        or a list of tuples:
        https://alembic.sqlalchemy.org/en/latest/api/autogenerate.html#getting-diffs
        """
        # TODO: (D Release) Remove this method and its usage
        def ignore_leftover_nested_quota(element):
            operation = element[0]
            if operation == 'remove_column':
                table, column = element[2], element[3].name
                return (table, column) in (('quotas', 'allocated'),
                                           ('reservations', 'allocated_id'))
            if operation == 'remove_index':
                return element[1].name == 'ix_reservations_allocated_id'
            if operation == 'remove_fk':
                return (element[1].table.name == 'reservations'
                        and element[1].column_keys == ['allocated_id'])
            return False

        def include_element(element):
            """Determine whether diff element should be excluded."""
            if element[0] == 'modify_nullable':
                table_name, column = element[2], element[3]
                return (table_name, column) not in {
                    # NOTE(stephenfin): This field has nullable=True set, but
                    # since it's also a primary key (primary_key=True) the
                    # resulting schema will still end up being "NOT NULL". This
                    # weird combination was deemed necessary because MySQL will
                    # otherwise set this to "NOT NULL DEFAULT ''" which may be
                    # harmless but is inconsistent with other models. See the
                    # migration for more information.
                    ('encryption', 'encryption_id'),
                    # NOTE(stephenfin): The nullability of these fields is
                    # dependent on the backend, for some reason
                    ('encryption', 'provider'),
                    ('encryption', 'control_location'),
                }

            if ignore_leftover_nested_quota(element):
                return False

            return True

        def filter_elements(diff_directive):
            """Return only the elements that should not be ignored.

            It may return None or [] when all elements from the directive have
            been filtered out.
            """
            if isinstance(diff_directive, list):
                return [element for element in diff_directive
                        if include_element(element)]

            if include_element(diff_directive):
                return diff_directive
            return None

        result = []
        for diff_directive in diff:
            remaining = filter_elements(diff_directive)
            if remaining:
                result.append(remaining)
        return result


class TestModelsSyncSQLite(
    CinderModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    pass


class TestModelsSyncMySQL(
    CinderModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestModelsSyncPostgreSQL(
    CinderModelsMigrationsSync,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class MigrationsWalk(
    test_fixtures.OpportunisticDBTestMixin, test_base.BaseTestCase,
):
    # Migrations can take a long time, particularly on underpowered CI nodes.
    # Give them some breathing room.
    TIMEOUT_SCALING_FACTOR = 4
    BOOL_TYPE = sqlalchemy.types.BOOLEAN

    # NOTE: List of migrations where we allow dropping/altring things.
    # Reviewers: DO NOT ALLOW THINGS TO BE ADDED HERE WITHOUT CARE, and make
    # sure that in the previous release there was nothing using that column,
    # not even an ORM model (unless the whole ORM model was not being used)
    # See prevent_drop_alter method docstring.
    DROP_ALTER_EXCEPTIONS = [
        # Drops and alters from initial migration have already been accepted
        '921e1a36b076',
        # Making shared_targets explicitly nullable (DB already allowed it)
        'c92a3e68beed',
        # Migration 89aa6f9639f9 doesn't fail because it's for a SQLAlquemy
        # internal table, and we only check Cinder's tables.

        # Increasing resource column max length to 300 is acceptable, since
        # it's a backward compatible change.
        'b8660621f1b9',
        # Making use_quota non-nullable is acceptable since on the last release
        # we added an online migration to set the value, but we also provide
        # a default on the OVO, the ORM, and the DB engine.
        '9ab1b092a404',
        # Removing allocated_id and allocated columns is acceptable now since
        # we stopped using them in the code on the previous release.
        # TODO: (D Release) Uncomment next line
        # 'afd7494d43b7',
    ]
    FORBIDDEN_METHODS = ('alembic.operations.Operations.alter_column',
                         'alembic.operations.Operations.drop_column',
                         'alembic.operations.Operations.drop_table',
                         'alembic.operations.BatchOperations.alter_column',
                         'alembic.operations.BatchOperations.drop_column')

    VARCHAR_TYPE = sqlalchemy.types.VARCHAR

    def setUp(self):
        super().setUp()
        self.engine = enginefacade.writer.get_engine()
        self.patch(api, 'get_engine', lambda: self.engine)
        self.config = migration._find_alembic_conf()
        self.init_version = '921e1a36b076'

    @prevent_drop_alter
    def _migrate_up(self, revision, connection):
        check_method = getattr(self, f'_check_{revision}', None)
        if revision != self.init_version:  # no tests for the initial revision
            self.assertIsNotNone(
                check_method,
                f"API DB Migration {revision} doesn't have a test; add one"
            )

        pre_upgrade = getattr(self, f'_pre_upgrade_{revision}', None)
        if pre_upgrade:
            pre_upgrade(connection)

        alembic_api.upgrade(self.config, revision)

        if check_method:
            check_method(connection)

    def test_single_base_revision(self):
        """Ensure we only have a single base revision.

        There's no good reason for us to have diverging history, so validate
        that only one base revision exists. This will prevent simple errors
        where people forget to specify the base revision. If this fail for your
        change, look for migrations that do not have a 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_bases()))

    def test_single_head_revision(self):
        """Ensure we only have a single head revision.

        There's no good reason for us to have diverging history, so validate
        that only one head revision exists. This will prevent merge conflicts
        adding additional head revision points. If this fail for your change,
        look for migrations with the same 'revises' line in them.
        """
        script = alembic_script.ScriptDirectory.from_config(self.config)
        self.assertEqual(1, len(script.get_heads()))

    def test_walk_versions(self):
        with self.engine.begin() as connection:
            self.config.attributes['connection'] = connection
            script = alembic_script.ScriptDirectory.from_config(self.config)
            revisions = list(script.walk_revisions())
            # Need revisions from older to newer so the walk works as intended
            revisions.reverse()
            for revision_script in revisions:
                self._migrate_up(revision_script.revision, connection)

    def test_db_version_alembic(self):
        migration.db_sync()
        head = alembic_script.ScriptDirectory.from_config(
            self.config,
        ).get_current_head()
        self.assertEqual(head, migration.db_version())

    def _pre_upgrade_c92a3e68beed(self, connection):
        """Test shared_targets is nullable."""
        table = db_utils.get_table(connection, 'volumes')
        self._previous_type = type(table.c.shared_targets.type)

    def _check_c92a3e68beed(self, connection):
        """Test shared_targets is nullable."""
        table = db_utils.get_table(connection, 'volumes')
        self.assertIn('shared_targets', table.c)
        # Type hasn't changed
        self.assertIsInstance(table.c.shared_targets.type, self._previous_type)
        # But it's nullable
        self.assertTrue(table.c.shared_targets.nullable)

    def _check_daa98075b90d(self, connection):
        """Test resources have indexes."""
        for table in ('groups', 'group_snapshots', 'volumes', 'snapshots',
                      'backups'):
            db_utils.index_exists(connection,
                                  table,
                                  f'{table}_deleted_project_id_idx')

        db_utils.index_exists(connection,
                              'volumes', 'volumes_deleted_host_idx')

    def _check_89aa6f9639f9(self, connection):
        # the table only existed on legacy deployments: there's no way to check
        # for its removal without creating it first, which is dumb
        pass

    def _pre_upgrade_b8660621f1b9(self, connection):
        """Test resource columns were limited to 255 chars before."""
        for table_name in ('quotas', 'quota_classes', 'reservations'):
            table = db_utils.get_table(connection, table_name)
            self.assertIn('resource', table.c)
            self.assertIsInstance(table.c.resource.type, self.VARCHAR_TYPE)
            self.assertEqual(255, table.c.resource.type.length)

    def _check_b8660621f1b9(self, connection):
        """Test resource columns can be up to 300 chars."""
        for table_name in ('quotas', 'quota_classes', 'reservations'):
            table = db_utils.get_table(connection, table_name)
            self.assertIn('resource', table.c)
            self.assertIsInstance(table.c.resource.type, self.VARCHAR_TYPE)
            self.assertEqual(300, table.c.resource.type.length)

    def _check_9ab1b092a404(self, connection):
        """Test use_quota is non-nullable."""
        volumes = db_utils.get_table(connection, 'volumes')
        self.assertFalse(volumes.c.use_quota.nullable)
        snapshots = db_utils.get_table(connection, 'snapshots')
        self.assertFalse(snapshots.c.use_quota.nullable)

    def _check_b7b88f50aab5(self, connection):
        """Test consistencygroups quota was removed."""
        quota_classes = db_utils.get_table(connection, 'quota_classes')
        res = connection.execute(
            quota_classes.select().where(
                sqlalchemy.and_(
                    quota_classes.c.resource == 'consistencygroups',
                    ~quota_classes.c.deleted,
                    quota_classes.c.class_name == 'default')
            )).all()
        self.assertListEqual([], res)

    def _check_9c74c1c6971f(self, connection):
        """Test backup related quota was added."""
        quota_classes = db_utils.get_table(connection, 'quota_classes')
        res = connection.execute(
            sqlalchemy.select(quota_classes.c.resource).where(
                sqlalchemy.and_(
                    quota_classes.c.resource.startswith('backup'),
                    ~quota_classes.c.deleted,
                    quota_classes.c.class_name == 'default')
            )).all()

        self.assertEqual(2, len(res))
        self.assertEqual({'backups', 'backup_gigabytes'},
                         {r[0] for r in res})

    # TODO: (D Release) Uncomment method _check_afd7494d43b7 and create a
    # migration with hash afd7494d43b7 using the following command:
    #   $ tox -e venv -- alembic -c cinder/db/alembic.ini revision \
    #     --rev-id afd7494d43b7  -m 'drop quota leftovers'
    # Then replace the upgrade method in file
    # cinder/db/migrations/versions/afd7494d43b7_drop_quota_leftovers.py with
    # the uncommented upgrade method below, removing the unused sqlalchemy
    # import and adding "from oslo_db.sqlalchemy import utils as db_utils"
    #
    # def _check_afd7494d43b7(self, connection):
    #     """Test drop allocated related columns."""
    #     reservations = db_utils.get_table(connection, 'reservations')
    #     self.assertNotIn('allocated_id', reservations.c)
    #     quotas = db_utils.get_table(connection, 'quotas')
    #     self.assertNotIn('allocated', quotas.c)
    #
    # def upgrade():
    #     connection = op.get_bind()
    #     # SQLite doesn't support dropping columns, so we use a workaround
    #     if connection.engine.name == 'sqlite':
    #         with op.batch_alter_table('reservations') as batch_op:
    #             batch_op.drop_index(op.f('ix_reservations_allocated_id'))
    #             batch_op.drop_column('allocated_id')
    #         with op.batch_alter_table('quotas') as batch_op:
    #             batch_op.drop_column('allocated')
    #
    #     else:
    #         # The foreign key is unnamed and Cinder doesn't set a naming
    #         # convention, so the name was decided by the DB engine on
    #         #  creation, and each engine uses a different convention, so we
    #         # find out the name.
    #         fk_name = db_utils.get_foreign_key_constraint_name(
    #             connection, 'reservations', 'allocated_id')
    #         op.drop_constraint(fk_name, 'reservations', type_='foreignkey')
    #
    #         # Find out the name of the index as well
    #         indexes_names = db_utils.get_indexes(connection, 'reservations')
    #         index_name = [idx['name']
    #                       for idx in indexes_names
    #                       if idx['column_names'] == ['allocated_id']]
    #         # There HAS to be an index, but just to be safe...
    #         if index_name:
    #             # Use op.f to indicate that the index name already has the
    #             # naming convention applied to it.
    #             op.drop_index(op.f(index_name[0]), table_name='reservations')
    #
    #         op.drop_column('reservations', 'allocated_id')
    #         op.drop_column('quotas', 'allocated')
    #


class TestMigrationsWalkSQLite(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    pass


class TestMigrationsWalkMySQL(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestMigrationsWalkPostgreSQL(
    MigrationsWalk,
    test_fixtures.OpportunisticDBTestMixin,
    test_base.BaseTestCase,
):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from oslo_utils import units as o_units

from cinder import context
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.volume.drivers.open_e.jovian_common import driver
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom

UUID_1 = '12345678-1234-1234-1234-000000000001'
UUID_2 = '12345678-1234-1234-1234-000000000002'
UUID_3 = '12345678-1234-1234-1234-000000000003'
UUID_4 = '12345678-1234-1234-1234-000000000004'

UUID_S1 = '12345678-1234-1234-1234-100000000001'
UUID_S2 = '12345678-1234-1234-1234-100000000002'
UUID_S3 = '12345678-1234-1234-1234-100000000003'
UUID_S4 = '12345678-1234-1234-1234-100000000004'

CONFIG_OK = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'false',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '128K'
}

CONFIG_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'false',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '64K'
}

CONFIG_BAD_BLOCK_SIZE = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'false',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '61K'
}

CONFIG_BACKEND_NAME = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

CONFIG_MULTI_HOST = {
    'san_hosts': ['192.168.0.2', '192.168.0.3'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'jovian_user': 'admin',
    'jovian_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'volume_backend_name': 'JovianDSS',
    'reserved_percentage': 10,
    'jovian_block_size': '128K'
}

VOLUME_GET_NO_SNAPSHOTS = {
    "origin": None,
    "relatime": None,
    "acltype": None,
    "vscan": None,
    "full_name": f"Pool-0/{jcom.vname(UUID_1)}",
    "userrefs": None,
    "primarycache": "all",
    "logbias": "latency",
    "creation": "1695048563",
    "sync": "always",
    "is_clone": False,
    "dedup": "off",
    "sharenfs": None,
    "receive_resume_token": None,
    "volsize": "1073741824",
    "referenced": "57344",
    "sharesmb": None,
    "createtxg": "19806101",
    "reservation": "0",
    "scontext": None,
    "mountpoint": None,
    "casesensitivity": None,
    "guid": "13628065397986503663",
    "usedbyrefreservation": "1079975936",
    "dnodesize": None,
    "written": "57344",
    "logicalused": "28672",
    "compressratio": "1.00",
    "rootcontext": "none",
    "default_scsi_id": "9e697f6e11336500",
    "type": "volume",
    "compression": "lz4",
    "snapdir": None,
    "overlay": None,
    "encryption": "off",
    "xattr": None,
    "volmode": "default",
    "copies": "1",
    "snapshot_limit": "18446744073709551615",
    "aclinherit": None,
    "defcontext": "none",
    "readonly": "off",
    "version": None,
    "recordsize": None,
    "filesystem_limit": None,
    "mounted": None,
    "mlslabel": "none",
    "secondarycache": "all",
    "refreservation": "1080033280",
    "available": "955831783424",
    "san:volume_id": "9e697f6e11336500480c13e4467b7964bed4b02e",
    "encryptionroot": None,
    "exec": None,
    "refquota": None,
    "refcompressratio": "1.00",
    "quota": None,
    "utf8only": None,
    "keylocation": "none",
    "snapdev": "hidden",
    "snapshot_count": "18446744073709551615",
    "fscontext": "none",
    "clones": None,
    "canmount": None,
    "keystatus": None,
    "atime": None,
    "usedbysnapshots": "0",
    "normalization": None,
    "usedbychildren": "0",
    "volblocksize": "65536",
    "usedbydataset": "57344",
    "objsetid": "18142",
    "name": "a1",
    "defer_destroy": None,
    "pbkdf2iters": "0",
    "checksum": "on",
    "redundant_metadata": "all",
    "filesystem_count": None,
    "devices": None,
    "keyformat": "none",
    "setuid": None,
    "used": "1080033280",
    "logicalreferenced": "28672",
    "context": "none",
    "zoned": None,
    "nbmand": None,
}

VOLUME_GET_THAT_IS_CLONE = {
    "origin": f"Pool-0/{jcom.vname(UUID_1)}@{jcom.sname(UUID_S1, UUID_1)}",
    "relatime": None,
    "acltype": None,
    "vscan": None,
    "full_name": f"Pool-0/{jcom.vname(UUID_2)}",
    "userrefs": None,
    "primarycache": "all",
    "logbias": "latency",
    "creation": "1695078560",
    "sync": "always",
    "is_clone": True,
    "dedup": "off",
    "sharenfs": None,
    "receive_resume_token": None,
    "volsize": "1073741824",
    "referenced": "57344",
    "sharesmb": None,
    "createtxg": "19812058",
    "reservation": "0",
    "scontext": None,
    "mountpoint": None,
    "casesensitivity": None,
    "guid": "4947994863040470005",
    "usedbyrefreservation": "0",
    "dnodesize": None,
    "written": "0",
    "logicalused": "0",
    "compressratio": "1.00",
    "rootcontext": "none",
    "default_scsi_id": "5c02d042ed8dbce2",
    "type": "volume",
    "compression": "lz4",
    "snapdir": None,
    "overlay": None,
    "encryption": "off",
    "xattr": None,
    "volmode": "default",
    "copies": "1",
    "snapshot_limit": "18446744073709551615",
    "aclinherit": None,
    "defcontext": "none",
    "readonly": "off",
    "version": None,
    "recordsize": None,
    "filesystem_limit": None,
    "mounted": None,
    "mlslabel": "none",
    "secondarycache": "all",
    "refreservation": "0",
    "available": "954751713280",
    "san:volume_id": "5c02d042ed8dbce2570c8d5dc276dd6a2431e138",
    "encryptionroot": None,
    "exec": None,
    "refquota": None,
    "refcompressratio": "1.00",
    "quota": None,
    "utf8only": None,
    "keylocation": "none",
    "snapdev": "hidden",
    "snapshot_count": "18446744073709551615",
    "fscontext": "none",
    "clones": None,
    "canmount": None,
    "keystatus": None,
    "atime": None,
    "usedbysnapshots": "0",
    "normalization": None,
    "usedbychildren": "0",
    "volblocksize": "65536",
    "usedbydataset": "0",
    "objsetid": "19228",
    "name": "a2",
    "defer_destroy": None,
    "pbkdf2iters": "0",
    "checksum": "on",
    "redundant_metadata": "all",
    "filesystem_count": None,
    "devices": None,
    "keyformat": "none",
    "setuid": None,
    "used": "0",
    "logicalreferenced": "28672",
    "context": "none",
    "zoned": None,
    "nbmand": None,
}

SNAPSHOTS_GET_NO_CLONES = [
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 16:49:33',
     'volsize': '1073741824',
     'createtxg': '18402390',
     'guid': '15554334551928551694',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '106843',
     'name': jcom.sname(UUID_S1, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'},
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 18:44:49',
     'volsize': '1073741824',
     'createtxg': '18403768',
     'guid': '18319280142829358721',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '107416',
     'name': jcom.sname(UUID_S2, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'}]

SNAPSHOTS_GET_INTERMEDIATE_SNAP = [
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 16:49:33',
     'volsize': '1073741824',
     'createtxg': '18402390',
     'guid': '15554334551928551694',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '106843',
     'name': jcom.vname(UUID_S1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'}]

SNAPSHOTS_GET_ONE_CLONE = [
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 16:49:33',
     'volsize': '1073741824',
     'createtxg': '18402390',
     'guid': '15554334551928551694',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '106843',
     'name': jcom.sname(UUID_S1, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'},
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 18:44:49',
     'volsize': '1073741824',
     'createtxg': '18403768',
     'guid': '18319280142829358721',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '107416',
     'clones': 'Pool-0/' + jcom.vname(UUID_2),
     'name': jcom.sname(UUID_S2, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'}]

SNAPSHOTS_GET_MULTIPLE_CLONES = [
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 16:49:33',
     'volsize': '1073741824',
     'createtxg': '18402390',
     'guid': '15554334551928551694',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '106843',
     'name': jcom.sname(UUID_S1, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'},
    {'referenced': '57344',
     'userrefs': '0',
     'primarycache': 'all',
     'creation': '2023-06-28 18:44:49',
     'volsize': '1073741824',
     'createtxg': '18403768',
     'guid': '18319280142829358721',
     'compressratio': '1.00',
     'rootcontext': 'none',
     'encryption': 'off',
     'defcontext': 'none',
     'written': '0',
     'type': 'snapshot',
     'secondarycache': 'all',
     'used': '0',
     'refcompressratio': '1.00',
     'fscontext': 'none',
     'objsetid': '107416',
     'clones': f'Pool-0/{jcom.vname(UUID_2)},Pool-0/{jcom.vname(UUID_3)}',
     'name': jcom.sname(UUID_S2, UUID_1),
     'defer_destroy': 'off',
     'san:volume_id': 'e82c7fcbd78df0ffe67d363412e5091421d313ca',
     'mlslabel': 'none',
     'logicalreferenced': '28672',
     'context': 'none'}]

SNAPSHOT_GET_ONE_CLONE = {
    "referenced": "57344",
    "userrefs": "0",
    "primarycache": "all",
    "creation": "2023-09-19 01:08:25",
    "volsize": "1073741824",
    "createtxg": "19812047",
    "guid": "7433980076067517643",
    "compressratio": "1.00",
    "rootcontext": "none",
    "encryption": "off",
    "defcontext": "none",
    "written": "57344",
    "type": "snapshot",
    "secondarycache": "all",
    "used": "0",
    "refcompressratio": "1.00",
    "fscontext": "none",
    "clones": f"Pool-0/{jcom.vname(UUID_2)}",
    "objsetid": "19220",
    "defer_destroy": "off",
    "san:volume_id": "9e697f6e11336500480c13e4467b7964bed4b02e",
    "mlslabel": "none",
    "logicalreferenced": "28672",
    "context": "none"
}

SNAPSHOTS_CASCADE_1 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_3)}]

SNAPSHOTS_CASCADE_2 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_3)}]

SNAPSHOTS_CASCADE_3 = [
    {"name": jcom.vname(UUID_4),
     "clones": "Pool-0/" + jcom.vname(UUID_4)}]

SNAPSHOTS_EMPTY = []

SNAPSHOTS_CLONE = [
    {"name": jcom.vname(UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_1)}]

SNAPSHOTS_GARBAGE = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.vname(UUID_2)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": ""}]

SNAPSHOTS_RECURSIVE_1 = [
    {"name": jcom.sname(UUID_S1, UUID_1),
     "clones": "Pool-0/" + jcom.sname(UUID_S1, UUID_1)},
    {"name": jcom.sname(UUID_S1, UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]

SNAPSHOTS_RECURSIVE_CHAIN_1 = [
    {"name": jcom.sname(UUID_S1, UUID_3),
     "clones": "Pool-0/" + jcom.hidden(UUID_3)}]

SNAPSHOTS_RECURSIVE_CHAIN_2 = [
    {"name": jcom.vname(UUID_2),
     "clones": "Pool-0/" + jcom.hidden(UUID_2)}]


def get_jdss_exceptions():

    out = [jexc.JDSSException(reason="Testing"),
           jexc.JDSSRESTException(request="ra request", reason="Testing"),
           jexc.JDSSRESTProxyException(host="test_host", reason="Testing"),
           jexc.JDSSResourceNotFoundException(res="test_resource"),
           jexc.JDSSVolumeNotFoundException(volume="test_volume"),
           jexc.JDSSSnapshotNotFoundException(snapshot="test_snapshot"),
           jexc.JDSSResourceExistsException(res="test_resource"),
           jexc.JDSSSnapshotExistsException(snapshot="test_snapshot"),
           jexc.JDSSVolumeExistsException(volume="test_volume"),
           jexc.JDSSSnapshotIsBusyException(snapshot="test_snapshot")]

    return out


class TestOpenEJovianDSSDriver(test.TestCase):

    def get_jdss_driver(self, config):
        ctx = context.get_admin_context()

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None
        cfg.get = lambda val, default: config.get(val, default)

        jdssd = driver.JovianDSSDriver(cfg)

        lib_to_patch = ('cinder.volume.drivers.open_e.jovian_common.driver.'
                        'rest.JovianRESTAPI')
        with mock.patch(lib_to_patch) as ra:
            ra.is_pool_exists.return_value = True
        jdssd.ra = mock.Mock()
        return jdssd, ctx

    def start_patches(self, patches):
        for p in patches:
            p.start()

    def stop_patches(self, patches):
        for p in patches:
            p.stop()

    def test_create_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vol = fake_volume.fake_volume_obj(ctx)

        vol.id = UUID_1
        vol.size = 1

        jdssd.ra.create_lun.return_value = None

        jdssd.create_volume(vol.id, 1)

        create_vol_expected = [mock.call(jcom.vname(vol.id),
                                         1073741824,
                                         sparse=False,
                                         block_size=None)]

        jdssd.create_volume(vol.id, 1, sparse=True)

        create_vol_expected += [mock.call(jcom.vname(vol.id),
                                          1073741824,
                                          sparse=True,
                                          block_size=None)]

        jdssd.create_volume(vol.id, 1, sparse=True, block_size="64K")

        create_vol_expected += [mock.call(jcom.vname(vol.id),
                                          1073741824,
                                          sparse=True,
                                          block_size="64K")]

        jdssd.ra.create_lun.assert_has_calls(create_vol_expected)

    def test_promote_newest_delete_no_snapshots(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        # test provide empty snapshot list
        snapshots = []
        jdssd.ra.get_snapshots.return_value = []
        resp = {'data': {"vscan": None,
                         "full_name": "Pool-0/v_" + UUID_1,
                         "userrefs": None,
                         "primarycache": "all",
                         "logbias": "latency",
                         "creation": "1591543140",
                         "sync": "always",
                         "is_clone": False,
                         "dedup": "off",
                         "sharenfs": None,
                         "receive_resume_token": None,
                         "volsize": "1073741824"},
                'error': None,
                'code': 200}
        jdssd.ra.get_lun.return_value = resp
        jdssd.ra.delete_lun.return_value = None
        jdssd._promote_newest_delete(vname, snapshots)
        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=True)]
        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)

        # test provide none as snapshot list
        snapshots = None
        jdssd.ra.get_snapshots.return_value = []
        resp = {'data': {"vscan": None,
                         "full_name": "Pool-0/v_" + UUID_1,
                         "userrefs": None,
                         "primarycache": "all",
                         "logbias": "latency",
                         "creation": "1591543140",
                         "sync": "always",
                         "is_clone": False,
                         "dedup": "off",
                         "sharenfs": None,
                         "receive_resume_token": None,
                         "volsize": "1073741824"},
                'error': None,
                'code': 200}
        jdssd.ra.get_lun.return_value = resp
        jdssd.ra.delete_lun.return_value = None
        jdssd._promote_newest_delete(vname, snapshots)
        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=True)]
        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)

    def test_promote_newest_delete_has_snapshots(self):
        '''Test promote-remove on volume with snapshots

        We should sucessevely remove volume if it have snapshots
        with no clones.
        Also no promote should be called.
        '''

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        snapshots = copy.deepcopy(SNAPSHOTS_GET_NO_CLONES)
        jdssd.ra.get_snapshots.return_value = []
        resp = {"vscan": None,
                "full_name": "Pool-0/v_" + UUID_1,
                "userrefs": None,
                "primarycache": "all",
                "logbias": "latency",
                "creation": "1591543140",
                "sync": "always",
                "is_clone": False,
                "dedup": "off",
                "sharenfs": None,
                "receive_resume_token": None,
                "volsize": "1073741824"},
        jdssd.ra.get_lun.return_value = resp
        jdssd.ra.delete_lun.return_value = None
        jdssd._promote_newest_delete(vname, snapshots)
        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=True)]
        jdssd.ra.promote.assert_not_called()
        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)

    def test_promote_newest_delete_has_clone(self):
        '''Test promote-remove on volume with clone

        We should sucessevely remove volume if it have snapshot
        with no clone.
        '''

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        snapshots = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)

        self.assertEqual(f'Pool-0/{jcom.vname(UUID_2)}',
                         snapshots[1]['clones'])
        jdssd.ra.get_snapshots.return_value = []
        resp = {'data': {"vscan": None,
                         "full_name": "Pool-0/v_" + UUID_1,
                         "userrefs": None,
                         "primarycache": "all",
                         "logbias": "latency",
                         "creation": "1591543140",
                         "sync": "always",
                         "is_clone": False,
                         "dedup": "off",
                         "sharenfs": None,
                         "receive_resume_token": None,
                         "volsize": "1073741824"},
                'error': None,
                'code': 200}
        jdssd.ra.get_lun.return_value = resp
        jdssd.ra.delete_lun.return_value = None

        jdssd._promote_newest_delete(vname, snapshots)
        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=True)]
        promote_vol_expected = [mock.call(vname,
                                          jcom.sname(UUID_S2, UUID_1),
                                          jcom.vname(UUID_2))]
        jdssd.ra.promote.assert_has_calls(promote_vol_expected)
        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)

    def test_promote_newest_delete_has_multiple_clones(self):
        '''Test promote-remove on volume with clone

        We should sucessevely remove volume if it have snapshot
        with no clone.
        '''

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        snapshots = copy.deepcopy(SNAPSHOTS_GET_MULTIPLE_CLONES)
        jdssd.ra.get_snapshots.return_value = []
        resp = {'data': {"vscan": None,
                         "full_name": "Pool-0/s_" + UUID_S2,
                         "userrefs": None,
                         "primarycache": "all",
                         "logbias": "latency",
                         "creation": "1591543140",
                         "sync": "always",
                         "is_clone": False,
                         "dedup": "off",
                         "sharenfs": None,
                         "receive_resume_token": None,
                         "volsize": "1073741824"},
                'error': None,
                'code': 200}
        jdssd.ra.get_lun.return_value = resp
        jdssd.ra.delete_lun.return_value = None
        jdssd._promote_newest_delete(vname, snapshots)
        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=True)]
        promote_vol_expected = [mock.call(vname,
                                          jcom.sname(UUID_S2, UUID_1),
                                          jcom.vname(UUID_3))]
        jdssd.ra.promote.assert_has_calls(promote_vol_expected)
        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)

    def test_delete_vol_with_source_snap_no_snap(self):
        '''Test _delete_vol_with_source_snap

        We should sucessevely remove volume with no snapshots.
        '''
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        jdssd.ra.get_lun.return_value = copy.deepcopy(VOLUME_GET_NO_SNAPSHOTS)
        jdssd.ra.delete_lun.return_value = None

        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=False)]

        jdssd._delete_vol_with_source_snap(vname, recursive=False)

        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)
        jdssd.ra.delete_snapshot.assert_not_called()

    def test_delete_vol_with_source_snap(self):
        '''Test _delete_vol_with_source_snap

        We should sucessevely remove volume that is clone.
        We should not remove source snapshot
        if that snapshot is not related to volume to remove
        '''
        # Snapshot does belong to parent volume

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_2)
        jdssd.ra.get_lun.return_value = \
            copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)
        jdssd.ra.delete_lun.return_value = None

        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=False)]

        jdssd._delete_vol_with_source_snap(vname, recursive=False)

        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)
        jdssd.ra.delete_snapshot.assert_not_called()

    def test_delete_vol_with_source_snap_snap_delete(self):
        # Snapshot belongs to volume to delete

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_2)
        lun_info = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)
        origin = f"Pool-0/{jcom.vname(UUID_1)}@{jcom.sname(UUID_S2, UUID_2)}"
        lun_info['origin'] = origin

        jdssd.ra.get_lun.return_value = lun_info

        jdssd.ra.delete_lun.return_value = None

        delete_vol_expected = [mock.call(vname,
                                         force_umount=True,
                                         recursively_children=False)]
        delete_snapshot_expected = [mock.call(jcom.vname(UUID_1),
                                              jcom.sname(UUID_S2, UUID_2),
                                              recursively_children=True,
                                              force_umount=True)]
        jdssd._delete_vol_with_source_snap(vname, recursive=False)

        jdssd.ra.delete_lun.assert_has_calls(delete_vol_expected)
        jdssd.ra.delete_snapshot.assert_has_calls(delete_snapshot_expected)

    def test_clean_garbage_resources(self):

        # Make sure that we request list of snapshots if none is provide
        # Make sure we remove intermediate volume like snapshot if it has
        # no volumes associated with it
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        snap_list = copy.deepcopy(SNAPSHOTS_GET_INTERMEDIATE_SNAP)
        get_snapshots_expectes = [mock.call(jcom.vname(UUID_1))]
        delete_snapshot_expected = [mock.call(jcom.vname(UUID_1),
                                              jcom.vname(UUID_S1),
                                              force_umount=True)]

        jdssd.ra.get_snapshots.side_effect = [snap_list, []]

        ret = jdssd._clean_garbage_resources(vname, snapshots=None)

        self.assertEqual([], ret)
        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots_expectes)
        jdssd.ra.delete_snapshot.assert_has_calls(delete_snapshot_expected)

    def test_clean_garbage_resources_do_nothing(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        snap_list = SNAPSHOTS_GET_ONE_CLONE.copy()
        get_snapshots_expectes = [mock.call(jcom.vname(UUID_1))]

        jdssd.ra.get_snapshots.side_effect = [snap_list, snap_list]

        ret = jdssd._clean_garbage_resources(vname, snapshots=None)

        self.assertEqual(SNAPSHOTS_GET_ONE_CLONE, ret)
        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots_expectes)
        jdssd.ra.delete_snapshot.assert_not_called()

    def test_clean_garbage_resources_clean_hidden(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        snap_list[1]['clones'] = f"Pool-0/{jcom.hidden(UUID_2)}"
        snap_list[1]['name'] = jcom.sname(UUID_S2, UUID_1)

        get_snapshots_expectes = [mock.call(jcom.vname(UUID_1)),
                                  mock.call(jcom.vname(UUID_1))]

        jdssd.ra.get_snapshots.side_effect = [snap_list,
                                              SNAPSHOTS_GET_NO_CLONES]
        with mock.patch.object(jdssd, '_promote_newest_delete') as pnd:
            pnd.side_effect = [None]
            ret = jdssd._clean_garbage_resources(vname, snapshots=None)
            self.assertEqual(SNAPSHOTS_GET_NO_CLONES, ret)
            pnd.assert_has_calls([mock.call(jcom.hidden(UUID_2))])

        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots_expectes)
        jdssd.ra.delete_snapshot.assert_not_called()

    def test_clean_garbage_resources_clean_snapshot(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        snap_list[1]['clones'] = f"Pool-0/{jcom.sname(UUID_S2, UUID_1)}"
        snap_list[1]['name'] = jcom.sname(UUID_S2, UUID_1)

        get_snapshots_expectes = [mock.call(jcom.vname(UUID_1))]

        jdssd.ra.get_snapshots.side_effect = [snap_list]
        with mock.patch.object(jdssd, '_promote_newest_delete') as pnd:
            pnd.side_effect = [None]

            ret = jdssd._clean_garbage_resources(vname, snapshots=None)
            pnd.assert_not_called()
            self.assertEqual(snap_list, ret)

        jdssd.ra.get_snapshots.assert_has_calls(get_snapshots_expectes)
        jdssd.ra.delete_snapshot.assert_not_called()

    def test_list_busy_snapshots(self):

        # Check operation with regular clone
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)

        ret = jdssd._list_busy_snapshots(vname, snap_list)
        self.assertEqual([SNAPSHOTS_GET_ONE_CLONE[1]], ret)

        # Check hidden clone
        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)

        snap_list[0]['clones'] = f"Pool-0/{jcom.hidden(UUID_2)}"
        snap_list[0]['name'] = jcom.sname(UUID_S2, UUID_1)

        ret = jdssd._list_busy_snapshots(vname, snap_list)
        self.assertEqual(snap_list, ret)

        # Check exlude dedicated volume flag
        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)

        ret = jdssd._list_busy_snapshots(vname, snap_list,
                                         exclude_dedicated_volumes=True)
        self.assertEqual([], ret)

    def _clean_volume_snapshots_mount_points(self):

        # Single attached snapshot case
        vname = jcom.vname(UUID_1)
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        cname = jcom.sname(UUID_S2, UUID_1)
        snap_list[1]['clones'] = cname = f'Pool-0/{cname}'

        ret_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        ret_list[1].pop('clones')

        jdssd.ra.get_snapshots.return_value = ret_list
        with mock.patch.object(jdssd, '_delete_volume'):
            ret = jdssd._clean_volume_snapshots_mount_points(vname, snap_list)
            jdssd._delete_volume.assert_called_once_with(cname, cascade=True)

        jdssd.ra.get_snapshots.assert_called_once_with(vname)
        self.assertEqual(ret_list, ret)

        # Multiple attached snapshot case
        vname = jcom.vname(UUID_1)
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        cname0 = jcom.sname(UUID_S1, UUID_1)
        cname1 = jcom.sname(UUID_S2, UUID_1)
        snap_list[0]['clones'] = cname = f'Pool-0/{cname0}'
        snap_list[1]['clones'] = cname = f'Pool-0/{cname1}'

        ret_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)
        ret_list[0].pop('clones')
        ret_list[1].pop('clones')

        jdssd.ra.get_snapshots.return_value = ret_list

        del_vol_expected = [mock.call(jcom.vname(cname0), cascade=True),
                            mock.call(jcom.vname(cname1), cascade=True)]

        with mock.patch.object(jdssd, '_delete_volume'):
            ret = jdssd._clean_volume_snapshots_mount_points(vname, snap_list)
            jdssd._delete_volume.assert_has_calls(del_vol_expected)

        jdssd.ra.get_snapshots.assert_called_once_with(vname)
        self.assertEqual(ret_list, ret)

    def test_delete_volume_no_snap(self):

        vname = jcom.vname(UUID_1)
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        jdssd.ra.delete_lun.return_value = None

        del_lun_exp = [mock.call(
            jcom.vname(UUID_1),
            force_umount=True,
            recursively_children=False)]

        jdssd._delete_volume(vname)

        jdssd.ra.delete_lun.assert_has_calls(del_lun_exp)
        jdssd.ra.get_snapshots.assert_not_called()

    def test_delete_volume_cascade_with_clones(self):

        vname = jcom.vname(UUID_1)
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        jdssd.ra.delete_lun.side_effect = [
            jexc.JDSSResourceIsBusyException(res=vname)]

        del_lun_exp = [mock.call(jcom.vname(UUID_1),
                                 force_umount=True,
                                 recursively_children=True)]

        snap_list = copy.deepcopy(SNAPSHOTS_GET_ONE_CLONE)

        get_snap_exp = [mock.call(jcom.vname(UUID_1))]

        jdssd.ra.get_snapshots.side_effect = [snap_list]

        pnd_exp = [mock.call(jcom.vname(UUID_1), snapshots=snap_list)]
        with mock.patch.object(jdssd, '_promote_newest_delete'):
            jdssd._delete_volume(vname, cascade=True)
            jdssd._promote_newest_delete.assert_has_calls(pnd_exp)

        jdssd.ra.delete_lun.assert_has_calls(del_lun_exp)
        jdssd.ra.get_snapshots.assert_has_calls(get_snap_exp)

    def test_delete_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        with mock.patch.object(jdssd, '_delete_volume'):
            jdssd.delete_volume(UUID_1)
            jdssd._delete_volume.assert_called_once_with(vname, cascade=False)

    def test_get_provider_location(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        host = CONFIG_OK["san_hosts"][0]
        port = CONFIG_OK["target_port"]
        target_name = CONFIG_OK["target_prefix"] + UUID_1
        patches = [mock.patch.object(
            jdssd.ra,
            "get_active_host",
            return_value=host)]
        out = '{host}:{port},1 {name} 0'.format(
            host=host,
            port=port,
            name=target_name
        )
        self.start_patches(patches)
        self.assertEqual(out, jdssd.get_provider_location(UUID_1))
        self.stop_patches(patches)

    def test_get_target_name(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        ret = jdssd._get_target_name(UUID_1)
        self.assertEqual(ret, f'iqn.2020-04.com.open-e.cinder:{UUID_1}')

    def test_get_iscsi_properties(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        provider_auth = 'chap user_name 123456789012'

        multipath = True

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        ret = jdssd._get_iscsi_properties(UUID_1,
                                          provider_auth,
                                          multipath=multipath)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqns': [target_name],
                    'target_lun': 0,
                    'target_luns': [0],
                    'target_portals': ['192.168.0.2:3260']}
        self.assertEqual(expected, ret)

    def test_get_iscsi_properties_multipath(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_MULTI_HOST)

        provider_auth = 'chap user_name 123456789012'

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        ret = jdssd._get_iscsi_properties(UUID_1,
                                          provider_auth,
                                          multipath=True)
        expected = {'auth_method': 'chap',
                    'auth_password': '123456789012',
                    'auth_username': 'user_name',
                    'target_discovered': False,
                    'target_iqns': [target_name, target_name],
                    'target_lun': 0,
                    'target_luns': [0, 0],
                    'target_portals': ['192.168.0.2:3260', '192.168.0.3:3260']}
        self.assertEqual(expected, ret)

    def test_remove_target_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._remove_target_volume(UUID_1, jcom.vname(UUID_1))

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_remove_target_volume_no_target(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.detach_target_vol.side_effect = (
            jexc.JDSSResourceNotFoundException(res=target_name))
        jdssd.ra.delete_target.return_value = None

        jdssd._remove_target_volume(UUID_1, vname)

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_remove_target_volume_fail_to_detach(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.side_effect = (
            jexc.JDSSRESTException(reason='running test', request='test'))
        jdssd.ra.delete_target.return_value = None

        self.assertRaises(jexc.JDSSException,
                          jdssd._remove_target_volume,
                          UUID_1,
                          jcom.vname(UUID_1))

        jdssd.ra.detach_target_vol.assert_called_once_with(
            target_name, jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_not_called()

    def test_remove_target_volume_fail_to_delete(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        jdssd.ra.detach_target_vol.return_value = None
        jdssd.ra.delete_target.side_effect = (
            jexc.JDSSRESTException(reason='running test', request='test'))

        self.assertRaises(jexc.JDSSException,
                          jdssd._remove_target_volume,
                          UUID_1,
                          jcom.vname(UUID_1))

        jdssd.ra.detach_target_vol.assert_called_once_with(target_name,
                                                           jcom.vname(UUID_1))
        jdssd.ra.delete_target.assert_called_with(target_name)

    def test_ensure_export(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_MULTI_HOST)

        provider_auth = 'chap user_name 123456789012'

        with mock.patch.object(jdssd, "_ensure_target_volume"):
            jdssd.ensure_export(UUID_1, provider_auth)
            jdssd._ensure_target_volume.assert_called_once_with(
                UUID_1,
                jcom.vname(UUID_1),
                provider_auth)

    def test_initialize_connection(self):

        # Test Ok
        jdssd, ctx = self.get_jdss_driver(CONFIG_MULTI_HOST)

        volume_id = UUID_1
        provider_auth = 'chap user_name 123456789012'

        multipath = True

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqns': [target_name, target_name],
                      'target_lun': 0,
                      'target_luns': [0, 0],
                      'target_portals': ['192.168.0.2:3260',
                                         '192.168.0.3:3260']}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

        vname = jcom.vname(volume_id)
        with mock.patch.object(jdssd, '_ensure_target_volume'):
            ret = jdssd.initialize_connection(volume_id, provider_auth,
                                              multipath=multipath)
            jdssd._ensure_target_volume.assert_called_with(UUID_1, vname,
                                                           provider_auth)

        self.assertEqual(con_info, ret)

        # Test initialize for snapshot
        jdssd, ctx = self.get_jdss_driver(CONFIG_MULTI_HOST)

        volume_id = UUID_1
        snapshot_id = UUID_S1

        provider_auth = 'chap user_name 123456789012'

        multipath = True

        target_name = CONFIG_OK['target_prefix'] + UUID_S1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqns': [target_name, target_name],
                      'target_lun': 0,
                      'target_luns': [0, 0],
                      'target_portals': ['192.168.0.2:3260',
                                         '192.168.0.3:3260']}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

        sname = jcom.sname(snapshot_id, volume_id)
        with mock.patch.object(jdssd, '_ensure_target_volume'):
            ret = jdssd.initialize_connection(volume_id, provider_auth,
                                              snapshot_id=snapshot_id,
                                              multipath=multipath)
            jdssd._ensure_target_volume.assert_called_with(UUID_S1, sname,
                                                           provider_auth,
                                                           mode='ro')

        self.assertEqual(con_info, ret)

        # Test no auth
        jdssd, ctx = self.get_jdss_driver(CONFIG_MULTI_HOST)

        volume_id = UUID_1

        provider_auth = None

        multipath = True

        target_name = CONFIG_OK['target_prefix'] + UUID_1

        properties = {'auth_method': 'chap',
                      'auth_password': '123456789012',
                      'auth_username': 'user_name',
                      'target_discovered': False,
                      'target_iqns': [target_name, target_name],
                      'target_lun': 0,
                      'target_luns': [0, 0],
                      'target_portals': ['192.168.0.2:3260',
                                         '192.168.0.3:3260']}

        con_info = {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

        sname = jcom.sname(snapshot_id, volume_id)
        with mock.patch.object(jdssd, '_ensure_target_volume'):

            self.assertRaises(jexc.JDSSException,
                              jdssd.initialize_connection,
                              volume_id,
                              provider_auth,
                              multipath=multipath)
            jdssd._ensure_target_volume.assert_not_called()

    def test_create_target_volume(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vid = jcom.vname(UUID_1)
        target_name = CONFIG_OK['target_prefix'] + UUID_1
        provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name', 'password': '123456789012'}

        patches = [
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials")]

        self.start_patches(patches)
        jdssd._create_target_volume(UUID_1, vid, provider_auth)
        jdssd.ra.create_target.assert_called_once_with(target_name,
                                                       use_chap=True)
        jdssd._attach_target_volume.assert_called_once_with(
            target_name, jcom.vname(UUID_1))
        jdssd._set_target_credentials.assert_called_once_with(
            target_name, cred)
        self.stop_patches(patches)

    def test_create_target_volume_for_snapshot_attachment(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vid = jcom.sname(UUID_S1, UUID_1)
        target_name = CONFIG_OK['target_prefix'] + UUID_S1
        provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name', 'password': '123456789012'}

        patches = [
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials")]

        self.start_patches(patches)
        jdssd._create_target_volume(UUID_S1, vid, provider_auth)
        jdssd.ra.create_target.assert_called_once_with(target_name,
                                                       use_chap=True)
        jdssd._attach_target_volume.assert_called_once_with(
            target_name, jcom.sname(UUID_S1, UUID_1))
        jdssd._set_target_credentials.assert_called_once_with(
            target_name, cred)
        self.stop_patches(patches)

    def test_attach_target_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_BACKEND_NAME)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)

        jdssd.ra.attach_target_vol.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._attach_target_volume(target_name, vname)

        jdssd.ra.attach_target_vol.assert_called_once_with(
            target_name, vname)
        jdssd.ra.delete_target.assert_not_called()

        ex = jexc.JDSSResourceExistsException(res=target_name)
        jdssd.ra.attach_target_vol.side_effect = ex

        self.assertRaises(jexc.JDSSException,
                          jdssd._attach_target_volume,
                          target_name,
                          vname)
        jdssd.ra.delete_target.assert_called_once_with(target_name)

    def test_set_target_credentials(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_BACKEND_NAME)

        target_name = CONFIG_BACKEND_NAME['target_prefix'] + UUID_1
        cred = {'name': 'user_name', 'password': '123456789012'}

        jdssd.ra.create_target_user.return_value = None
        jdssd.ra.delete_target.return_value = None

        jdssd._set_target_credentials(target_name, cred)

        jdssd.ra.create_target_user.assert_called_once_with(
            target_name, cred)
        jdssd.ra.delete_target.assert_not_called()

        ex = jexc.JDSSResourceExistsException(res=target_name)
        jdssd.ra.create_target_user.side_effect = ex

        self.assertRaises(jexc.JDSSException,
                          jdssd._set_target_credentials,
                          target_name,
                          cred)
        jdssd.ra.delete_target.assert_called_once_with(target_name)

    def test_clone_object(self):

        # test ok
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        ovname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        cvname = jcom.vname(UUID_2)
        jdssd._clone_object(cvname, sname, ovname, sparse=True)

        jdssd.ra.create_snapshot.assesrt_not_called()
        jdssd.ra.create_volume_from_snapshot.assert_called_once_with(
            cvname, sname, ovname, sparse=True)

        # test create snapshot
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        ovname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        cvname = jcom.vname(UUID_2)
        jdssd._clone_object(cvname, sname, ovname, sparse=True)

        jdssd.ra.create_snapshot.assesrt_not_called()
        jdssd.ra.create_volume_from_snapshot.assert_called_once_with(
            cvname, sname, ovname, sparse=True)

        # test create from snapshot failed
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        ovname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        cvname = jcom.vname(UUID_2)
        jdssd.ra.create_volume_from_snapshot.side_effect = [
            jexc.JDSSVolumeExistsException(volume=cvname)]
        self.assertRaises(jexc.JDSSException,
                          jdssd._clone_object,
                          cvname,
                          sname,
                          ovname,
                          sparse=True)

        jdssd.ra.create_snapshot.assesrt_not_called()
        jdssd.ra.create_volume_from_snapshot.assert_called_once_with(
            cvname, sname, ovname, sparse=True)
        jdssd.ra.delete_snapshot(ovname, sname,
                                 force_umount=True,
                                 recursively_children=True)

    def test_resize_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vname = jcom.vname(UUID_1)
        jdssd.resize_volume(UUID_1, 2)

        jdssd.ra.extend_lun.assert_called_once_with(vname, o_units.Gi * 2)

    def test_create_cloned_volume(self):

        # test ok
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        cvname = jcom.vname(UUID_2)
        vname = jcom.vname(UUID_1)

        jdssd.ra.get_lun.return_value = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)
        with mock.patch.object(jdssd, '_clone_object'):
            jdssd.create_cloned_volume(UUID_2, UUID_1, 1, sparse=False)

            jdssd._clone_object.assert_called_once_with(
                cvname, cvname, vname,
                sparse=False,
                create_snapshot=True)

        # test clone from snapshot
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        cvname = jcom.vname(UUID_2)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        jdssd.ra.get_lun.return_value = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)
        with mock.patch.object(jdssd, '_clone_object'):
            jdssd.create_cloned_volume(UUID_2, UUID_1, 1,
                                       snapshot_name=UUID_S1,
                                       sparse=False)

            jdssd._clone_object.assert_called_once_with(
                cvname, sname, vname,
                sparse=False,
                create_snapshot=False)

        # test extend
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        cvname = jcom.vname(UUID_2)
        vname = jcom.vname(UUID_1)

        get_vol = copy.deepcopy(VOLUME_GET_THAT_IS_CLONE)

        get_vol['volsize'] = "1073145824"
        jdssd.ra.get_lun.return_value = get_vol
        with mock.patch.object(jdssd, '_clone_object'), \
                mock.patch.object(jdssd, "resize_volume"):
            jdssd.create_cloned_volume(UUID_2, UUID_1, 1, sparse=False)

            jdssd._clone_object.assert_called_once_with(
                cvname, cvname, vname,
                sparse=False,
                create_snapshot=True)
            jdssd.resize_volume.assert_called_once_with(UUID_2, 1)

    def test_create_snapshot(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_BACKEND_NAME)

        jdssd.create_snapshot(UUID_S1, UUID_1)

        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)

        jdssd.ra.create_snapshot.assert_called_once_with(vname, sname)

    def test_create_export_snapshot(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_BACKEND_NAME)
        provider_auth = 'chap user_name 123456789012'

        sname = jcom.sname(UUID_S1, UUID_1)
        vname = jcom.vname(UUID_1)

        with mock.patch.object(jdssd, '_clone_object'), \
                mock.patch.object(jdssd, '_ensure_target_volume'):
            jdssd.create_export_snapshot(UUID_S1, UUID_1, provider_auth)

            jdssd._clone_object.assert_called_once_with(sname, sname, vname,
                                                        sparse=True,
                                                        create_snapshot=False)

            jdssd._ensure_target_volume(UUID_S1, sname, provider_auth)

    def test_remove_export(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        patches = [
            mock.patch.object(
                jdssd,
                "_remove_target_volume",
                return_value=None)]

        self.start_patches(patches)

        jdssd.remove_export(UUID_1)
        jdssd._remove_target_volume.assert_called_once_with(UUID_1, vname)

        self.stop_patches(patches)

    def test_remove_export_snapshot(self):

        # remove ok
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        with mock.patch.object(jdssd, "_remove_target_volume"), \
                mock.patch.object(jdssd, "_delete_volume"):

            jdssd.remove_export_snapshot(UUID_S1, UUID_1)
            jdssd._delete_volume.assert_called_once()
            jdssd._remove_target_volume.assert_called_once()

        # remove export failed
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vname = jcom.sname(UUID_S1, UUID_1)
        with mock.patch.object(jdssd, "_remove_target_volume"), \
                mock.patch.object(jdssd, "_delete_volume"):

            jdssd._remove_target_volume.side_effect = [
                jexc.JDSSResourceIsBusyException(res=vname)]

            self.assertRaises(jexc.JDSSResourceIsBusyException,
                              jdssd.remove_export_snapshot,
                              UUID_S1,
                              UUID_1)
            jdssd._delete_volume.assert_called_once()
            jdssd._remove_target_volume.assert_called_once()

    def test_delete_snapshot(self):

        # Delete ok, letion of snapshot with no clones
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)

        jdssd.ra.delete_snapshot.side_effect = [None]

        jdssd._delete_snapshot(vname, sname)
        jdssd.ra.delete_snapshot.assert_called_once_with(vname, sname,
                                                         force_umount=True)

        # Test deletion of snapshot with clones
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        side_eff = [jexc.JDSSSnapshotIsBusyException(snapshot=sname), None]
        jdssd.ra.delete_snapshot.side_effect = side_eff

        side_eff = [copy.deepcopy(SNAPSHOT_GET_ONE_CLONE)]

        jdssd.ra.get_snapshot.side_effect = side_eff

        with mock.patch.object(jdssd, '_promote_newest_delete'):
            jdssd._delete_snapshot(vname, sname)
            jdssd._promote_newest_delete.assert_not_called()
        jdssd.ra.delete_snapshot.assert_called_once_with(vname, sname,
                                                         force_umount=True)

        # Test deletion of attached snapshot
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_S1, UUID_1)
        side_eff = [jexc.JDSSSnapshotIsBusyException(snapshot=sname), None]
        jdssd.ra.delete_snapshot.side_effect = side_eff

        get_snap = copy.deepcopy(SNAPSHOT_GET_ONE_CLONE)

        get_snap['clones'] = f"Pool-0/{sname}"
        side_eff = [get_snap]

        jdssd.ra.get_snapshot.side_effect = side_eff

        delete_snap_expected = [mock.call(vname, sname, force_umount=True),
                                mock.call(vname, sname, force_umount=True)]

        with mock.patch.object(jdssd, '_promote_newest_delete'):

            jdssd._delete_snapshot(vname, sname)
            jdssd._promote_newest_delete.assert_called_once_with(sname)

        jdssd.ra.delete_snapshot.assert_has_calls(delete_snap_expected)

    def test_delete_snapshot_wrapper(self):
        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        volume_name = UUID_1
        snapshot_name = UUID_S1

        with mock.patch.object(jdssd, "_delete_snapshot"):
            jdssd.delete_snapshot(volume_name, snapshot_name)

            jdssd._delete_snapshot.assert_called_once_with(
                jcom.vname(UUID_1),
                jcom.sname(UUID_S1, UUID_1))

    def test_ensure_target_volume(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        id = UUID_1
        vid = jcom.vname(UUID_1)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = True
        jdssd.ra.is_target_lun.return_value = True
        jdssd.ra.get_target_user.return_value = [cred]

        self.start_patches(patches)

        jdssd._ensure_target_volume(id, vid, provider_auth)

        jdssd.ra.is_target.assert_called_once_with(target_name)

        jdssd.ra.is_target_lun.assert_called_once_with(target_name, vid)

        jdssd.ra.get_target_user.assert_called_once_with(target_name)

        jdssd.ra.delete_target_user.assert_not_called()
        jdssd._set_target_credentials.assert_not_called()
        self.stop_patches(patches)

    def test_ensure_target_volume_not_attached(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        id = UUID_1
        vid = jcom.vname(UUID_1)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = True
        jdssd.ra.is_target_lun.return_value = False
        jdssd.ra.get_target_user.return_value = [cred]

        self.start_patches(patches)

        jdssd._ensure_target_volume(id, vid, provider_auth)

        jdssd.ra.is_target.assert_called_once_with(target_name)
        jdssd.ra.is_target_lun.assert_called_once_with(target_name, vid)

        jdssd._attach_target_volume.assert_called_once_with(
            target_name, vid)
        jdssd.ra.get_target_user.assert_called_once_with(target_name)

        jdssd.ra.delete_target_user.assert_not_called()
        jdssd._set_target_credentials.assert_not_called()
        self.stop_patches(patches)

    def test_ensure_target_volume_no_target(self):

        jdssd, ctx = self.get_jdss_driver(CONFIG_OK)

        id = UUID_1
        vid = jcom.vname(UUID_1)

        target_name = CONFIG_OK['target_prefix'] + UUID_1
        provider_auth = 'chap user_name 123456789012'

        cred = {'name': 'user_name'}

        patches = [
            mock.patch.object(jdssd, "_create_target_volume"),
            mock.patch.object(jdssd, "_attach_target_volume"),
            mock.patch.object(jdssd, "_set_target_credentials"),
            mock.patch.object(jdssd, "_attach_target_volume")]

        jdssd.ra.is_target.return_value = False
        jdssd.ra.get_target_user.return_value = cred['name']

        self.start_patches(patches)

        jdssd._ensure_target_volume(id, vid, provider_auth)

        jdssd.ra.is_target.assert_called_once_with(target_name)
        jdssd._create_target_volume.assert_called_once_with(id,
                                                            vid,
                                                            provider_auth)

        jdssd.ra.is_target_lun.assert_not_called()
        self.stop_patches(patches)
